"""Test message patching for internationalization purposes.

Runs the text builder in the test root.
"""

from __future__ import annotations

import os
import re
import shutil
import time
from typing import TYPE_CHECKING

import pygments
import pytest
from babel.messages import mofile, pofile
from babel.messages.catalog import Catalog
from docutils import nodes

from sphinx import locale
from sphinx._cli.util.errors import strip_escape_sequences
from sphinx.testing.util import assert_node, etree_parse
from sphinx.util.nodes import NodeMatcher

if TYPE_CHECKING:
    from collections.abc import Iterator
    from io import StringIO
    from pathlib import Path

_CATALOG_LOCALE = 'xx'

sphinx_intl = pytest.mark.sphinx(
    confoverrides={
        'language': _CATALOG_LOCALE,
        'locale_dirs': ['.'],
        'gettext_compact': False,
        'html_sidebars': {'**': ['globaltoc.html']},  # for test_html_meta
    },
)


def read_po(pathname):
    with open(pathname, encoding='utf-8') as f:
        return pofile.read_po(f)


def write_mo(pathname, po):
    with open(pathname, 'wb') as f:
        return mofile.write_mo(f, po)


def _set_mtime_ns(target: Path, value: int) -> int:
    os.utime(target, ns=(value, value))
    return target.stat().st_mtime_ns


def _get_bom_intl_path(app):
    basedir = app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES'
    return basedir / 'bom.po', basedir / 'bom.mo'


def _get_update_targets(app):
    app.env.find_files(app.config, app.builder)
    added, changed, removed = app.env.get_outdated_files(config_changed=False)
    return added, changed, removed


@pytest.fixture(autouse=True)
def _info(app):
    yield
    print('# language:', app.config.language)
    print('# locale_dirs:', app.config.locale_dirs)


def elem_gettexts(elem):
    return list(filter(None, map(str.strip, elem.itertext())))


def elem_getref(elem):
    return elem.attrib.get('refid') or elem.attrib.get('refuri')


def assert_elem(elem, texts=None, refs=None, names=None):
    if texts is not None:
        _texts = elem_gettexts(elem)
        assert _texts == texts
    if refs is not None:
        _refs = [elem_getref(x) for x in elem.findall('reference')]
        assert _refs == refs
    if names is not None:
        _names = elem.attrib.get('names').split()
        assert _names == names


def assert_count(expected_expr, result, count):
    find_pair = (expected_expr, result)
    assert len(re.findall(*find_pair)) == count, find_pair


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_emit_warnings(app):
    app.build()
    # test warnings in translation
    warnings = getwarning(app.warning)
    warning_expr = (
        '.*/warnings.txt:4:<translated>:1: '
        'WARNING: Inline literal start-string without end-string. \\[docutils\\]\n'
    )
    assert re.search(warning_expr, warnings), (
        f'{warning_expr!r} did not match {warnings!r}'
    )


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_warning_node(app):
    app.build()
    # test warnings in translation
    result = (app.outdir / 'warnings.txt').read_text(encoding='utf8')
    expect = (
        '3. I18N WITH REST WARNINGS'
        '\n**************************\n'
        '\nLINE OF >>``<<BROKEN LITERAL MARKUP.\n'
    )
    assert result == expect


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_title_underline(app):
    app.build()
    # --- simple translation; check title underlines
    result = (app.outdir / 'bom.txt').read_text(encoding='utf8')
    expect = (
        '2. Datei mit UTF-8'
        '\n******************\n'  # underline matches new translation
        '\nThis file has umlauts: äöü.\n'
    )
    assert result == expect


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_subdirs(app):
    app.build()
    # --- check translation in subdirs
    result = (app.outdir / 'subdir' / 'index.txt').read_text(encoding='utf8')
    assert result.startswith('1. subdir contents\n******************\n')


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_inconsistency_warnings(app):
    app.build()
    # --- check warnings for inconsistency in number of references
    result = (app.outdir / 'refs_inconsistency.txt').read_text(encoding='utf8')
    expect = (
        '8. I18N WITH REFS INCONSISTENCY'
        '\n*******************************\n'
        '\n* FOR CITATION [ref3].\n'
        '\n* reference FOR reference.\n'
        '\n* ORPHAN REFERENCE: I18N WITH REFS INCONSISTENCY.\n'
        '\n[1] THIS IS A AUTO NUMBERED FOOTNOTE.\n'
        '\n[ref2] THIS IS A CITATION.\n'
        '\n[100] THIS IS A NUMBERED FOOTNOTE.\n'
    )
    assert result == expect

    warnings = getwarning(app.warning)
    warning_fmt = (
        '.*/refs_inconsistency.txt:\\d+: '
        'WARNING: inconsistent %(reftype)s in translated message.'
        ' original: %(original)s, translated: %(translated)s \\[i18n.inconsistent_references\\]\n'
    )
    expected_warning_expr = (
        warning_fmt
        % {
            'reftype': 'footnote references',
            'original': "\\['\\[#\\]_'\\]",
            'translated': '\\[\\]',
        }
        + warning_fmt
        % {
            'reftype': 'footnote references',
            'original': "\\['\\[100\\]_'\\]",
            'translated': '\\[\\]',
        }
        + warning_fmt
        % {
            'reftype': 'references',
            'original': "\\['reference_'\\]",
            'translated': "\\['reference_', 'reference_'\\]",
        }
        + warning_fmt
        % {
            'reftype': 'references',
            'original': '\\[\\]',
            'translated': "\\['`I18N WITH REFS INCONSISTENCY`_'\\]",
        }
    )
    assert re.search(expected_warning_expr, warnings), (
        f'{expected_warning_expr!r} did not match {warnings!r}'
    )

    expected_citation_ref_warning_expr = '.*/refs_inconsistency.txt:\\d+: WARNING: Citation \\[ref2\\] is not referenced.'
    assert re.search(expected_citation_ref_warning_expr, warnings), (
        f'{expected_citation_ref_warning_expr!r} did not match {warnings!r}'
    )

    expected_citation_warning_expr = (
        '.*/refs_inconsistency.txt:\\d+: WARNING: citation not found: ref3'
    )
    assert re.search(expected_citation_warning_expr, warnings), (
        f'{expected_citation_warning_expr!r} did not match {warnings!r}'
    )


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_noqa(app):
    app.build()
    result = (app.outdir / 'noqa.txt').read_text(encoding='utf8')
    expect = r"""FIRST SECTION
*************

TRANSLATED TEXT WITHOUT REFERENCE.

TEST noqa WHITESPACE INSENSITIVITY.

"#noqa" IS ESCAPED AT THE END OF THIS STRING. #noqa


NEXT SECTION WITH PARAGRAPH TO TEST BARE noqa
*********************************************

Some text, again referring to the section: NEXT SECTION WITH PARAGRAPH
TO TEST BARE noqa.
"""
    assert result == expect
    assert 'next-section' not in getwarning(app.warning)


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_literalblock_warnings(app):
    app.build()
    # --- check warning for literal block
    result = (app.outdir / 'literalblock.txt').read_text(encoding='utf8')
    expect = (
        '9. I18N WITH LITERAL BLOCK'
        '\n**************************\n'
        '\nCORRECT LITERAL BLOCK:\n'
        '\n   this is'
        '\n   literal block\n'
        '\nMISSING LITERAL BLOCK:\n'
        '\n<SYSTEM MESSAGE:'
    )
    assert result.startswith(expect)

    warnings = getwarning(app.warning)
    expected_warning_expr = (
        '.*/literalblock.txt:\\d+: WARNING: Literal block expected; none found.'
    )
    assert re.search(expected_warning_expr, warnings), (
        f'{expected_warning_expr!r} did not match {warnings!r}'
    )


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_definition_terms(app):
    app.build()
    # --- definition terms: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/975,
    # https://github.com/sphinx-doc/sphinx/issues/2198, and
    # https://github.com/sphinx-doc/sphinx/issues/2205
    result = (app.outdir / 'definition_terms.txt').read_text(encoding='utf8')
    expect = (
        '13. I18N WITH DEFINITION TERMS'
        '\n******************************\n'
        '\nSOME TERM'
        '\n   THE CORRESPONDING DEFINITION\n'
        '\nSOME *TERM* WITH LINK'
        '\n   THE CORRESPONDING DEFINITION #2\n'
        '\nSOME **TERM** WITH : CLASSIFIER1 : CLASSIFIER2'
        '\n   THE CORRESPONDING DEFINITION\n'
        '\nSOME TERM WITH : CLASSIFIER[]'
        '\n   THE CORRESPONDING DEFINITION\n'
    )
    assert result == expect


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_glossary_term(app):
    app.build()
    # --- glossary terms: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/1090
    result = (app.outdir / 'glossary_terms.txt').read_text(encoding='utf8')
    expect = r"""18. I18N WITH GLOSSARY TERMS
****************************

SOME NEW TERM
   THE CORRESPONDING GLOSSARY

SOME OTHER NEW TERM
   THE CORRESPONDING GLOSSARY #2

LINK TO *SOME NEW TERM*.

TRANSLATED GLOSSARY SHOULD BE SORTED BY TRANSLATED TERMS:

TRANSLATED TERM XXX
   DEFINE XXX

TRANSLATED TERM YYY
   DEFINE YYY

TRANSLATED TERM ZZZ
VVV
   DEFINE ZZZ
"""
    assert result == expect
    warnings = getwarning(app.warning)
    assert warnings.count('term not in glossary') == 1


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_glossary_term_inconsistencies(app):
    app.build()
    # --- glossary term inconsistencies: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/1090
    result = (app.outdir / 'glossary_terms_inconsistency.txt').read_text(
        encoding='utf8'
    )
    expect = (
        '19. I18N WITH GLOSSARY TERMS INCONSISTENCY'
        '\n******************************************\n'
        '\n1. LINK TO *SOME NEW TERM*.\n'
        '\n2. LINK TO *TERM NOT IN GLOSSARY*.\n'
    )
    assert result == expect

    warnings = getwarning(app.warning)
    expected_warning_expr = (
        '.*/glossary_terms_inconsistency.txt:\\d+: '
        'WARNING: inconsistent term references in translated message.'
        " original: \\[':term:`Some term`', ':term:`Some other term`'\\],"
        " translated: \\[':term:`SOME NEW TERM`'\\] \\[i18n.inconsistent_references\\]\n"
    )
    assert re.search(expected_warning_expr, warnings), (
        f'{expected_warning_expr!r} did not match {warnings!r}'
    )
    expected_warning_expr = (
        '.*/glossary_terms_inconsistency.txt:\\d+:<translated>:1: '
        "WARNING: term not in glossary: 'TERM NOT IN GLOSSARY'"
    )
    assert re.search(expected_warning_expr, warnings), (
        f'{expected_warning_expr!r} did not match {warnings!r}'
    )


@sphinx_intl
@pytest.mark.sphinx('gettext', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_gettext')
def test_gettext_section(app):
    app.build()
    # --- section
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'section.po')
    actual = read_po(app.outdir / 'section.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.id in actual_msg_ids


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_section(app):
    app.build()
    # --- section
    result = (app.outdir / 'section.txt').read_text(encoding='utf8')
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'section.po')
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.string in result


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_seealso(app):
    app.build()
    # --- seealso
    result = (app.outdir / 'seealso.txt').read_text(encoding='utf8')
    expect = (
        '12. I18N WITH SEEALSO'
        '\n*********************\n'
        '\nSee also: SHORT TEXT 1\n'
        '\nSee also: LONG TEXT 1\n'
        '\nSee also:\n'
        '\n  SHORT TEXT 2\n'
        '\n  LONG TEXT 2\n'
    )
    assert result == expect


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_figure_captions(app):
    app.build()
    # --- figure captions: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/940
    result = (app.outdir / 'figure.txt').read_text(encoding='utf8')
    expect = (
        '14. I18N WITH FIGURE CAPTION'
        '\n****************************\n'
        '\n   [image]MY CAPTION OF THE FIGURE\n'
        '\n   MY DESCRIPTION PARAGRAPH1 OF THE FIGURE.\n'
        '\n   MY DESCRIPTION PARAGRAPH2 OF THE FIGURE.\n'
        '\n'
        '\n14.1. FIGURE IN THE BLOCK'
        '\n=========================\n'
        '\nBLOCK\n'
        '\n      [image]MY CAPTION OF THE FIGURE\n'
        '\n      MY DESCRIPTION PARAGRAPH1 OF THE FIGURE.\n'
        '\n      MY DESCRIPTION PARAGRAPH2 OF THE FIGURE.\n'
        '\n'
        '\n'
        '14.2. IMAGE URL AND ALT\n'
        '=======================\n'
        '\n'
        '[image: I18N -> IMG][image]\n'
        '\n'
        '   [image: IMG -> I18N][image]\n'
        '\n'
        '\n'
        '14.3. IMAGE ON SUBSTITUTION\n'
        '===========================\n'
        '\n'
        '\n'
        '14.4. IMAGE UNDER NOTE\n'
        '======================\n'
        '\n'
        'Note:\n'
        '\n'
        '  [image: i18n under note][image]\n'
        '\n'
        '     [image: img under note][image]\n'
    )
    assert result == expect


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_rubric(app):
    app.build()
    # --- rubric: regression test for pull request
    # https://github.com/sphinx-doc/sphinx/issues/190
    result = (app.outdir / 'rubric.txt').read_text(encoding='utf8')
    expect = (
        'I18N WITH RUBRIC'
        '\n****************\n'
        '\n-[ RUBRIC TITLE ]-\n'
        '\n'
        '\nRUBRIC IN THE BLOCK'
        '\n===================\n'
        '\nBLOCK\n'
        '\n   -[ RUBRIC TITLE ]-\n'
    )
    assert result == expect


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_docfields(app):
    app.build()
    # --- docfields
    result = (app.outdir / 'docfields.txt').read_text(encoding='utf8')
    expect = (
        '21. I18N WITH DOCFIELDS'
        '\n***********************\n'
        '\nclass Cls1\n'
        '\n   Parameters:'
        '\n      **param** -- DESCRIPTION OF PARAMETER param\n'
        '\nclass Cls2\n'
        '\n   Parameters:'
        '\n      * **foo** -- DESCRIPTION OF PARAMETER foo\n'
        '\n      * **bar** -- DESCRIPTION OF PARAMETER bar\n'
        '\nclass Cls3(values)\n'
        '\n   Raises:'
        '\n      **ValueError** -- IF THE VALUES ARE OUT OF RANGE\n'
        '\nclass Cls4(values)\n'
        '\n   Raises:'
        '\n      * **TypeError** -- IF THE VALUES ARE NOT VALID\n'
        '\n      * **ValueError** -- IF THE VALUES ARE OUT OF RANGE\n'
        '\nclass Cls5\n'
        '\n   Returns:'
        '\n      A NEW "Cls3" INSTANCE\n'
    )
    assert result == expect


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_admonitions(app):
    app.build()
    # --- admonitions
    # https://github.com/sphinx-doc/sphinx/issues/1206:
    # gettext did not translate admonition directive's title
    # see https://docutils.sourceforge.io/docs/ref/rst/directives.html#admonitions
    result = (app.outdir / 'admonitions.txt').read_text(encoding='utf8')
    directives = (
        'attention',
        'caution',
        'danger',
        'error',
        'hint',
        'important',
        'note',
        'tip',
        'warning',
        'admonition',
    )
    for d in directives:
        assert d.upper() + ' TITLE' in result
        assert d.upper() + ' BODY' in result

    # https://github.com/sphinx-doc/sphinx/issues/4938
    # `1. ` prefixed admonition title
    assert '1. ADMONITION TITLE' in result


@sphinx_intl
@pytest.mark.sphinx('gettext', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_gettext')
def test_gettext_toctree(app):
    app.build()
    # --- toctree (index.rst)
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'index.po')
    actual = read_po(app.outdir / 'index.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.id in actual_msg_ids
    # --- toctree (toctree.rst)
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'toctree.po')
    actual = read_po(app.outdir / 'toctree.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.id in actual_msg_ids


@sphinx_intl
@pytest.mark.sphinx('gettext', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_gettext')
def test_gettext_table(app):
    app.build()
    # --- toctree
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'table.po')
    actual = read_po(app.outdir / 'table.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.id in actual_msg_ids


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_table(app):
    app.build()
    # --- toctree
    result = (app.outdir / 'table.txt').read_text(encoding='utf8')
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'table.po')
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.string in result


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_toctree(app):
    app.build()
    # --- toctree (index.rst)
    # Note: index.rst contains contents that is not shown in text.
    result = (app.outdir / 'index.txt').read_text(encoding='utf8')
    assert 'CONTENTS' in result
    assert 'TABLE OF CONTENTS' in result
    # --- toctree (toctree.rst)
    result = (app.outdir / 'toctree.txt').read_text(encoding='utf8')
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'toctree.po')
    for expect_msg in (m for m in expect if m.id):
        assert expect_msg.string in result


@sphinx_intl
@pytest.mark.sphinx('gettext', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_gettext')
def test_gettext_topic(app):
    app.build()
    # --- topic
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'topic.po')
    actual = read_po(app.outdir / 'topic.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.id in actual_msg_ids


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_topic(app):
    app.build()
    # --- topic
    result = (app.outdir / 'topic.txt').read_text(encoding='utf8')
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'topic.po')
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.string in result


@sphinx_intl
@pytest.mark.sphinx('gettext', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_gettext')
def test_gettext_definition_terms(app):
    app.build()
    # --- definition terms: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/2198,
    # https://github.com/sphinx-doc/sphinx/issues/2205
    expect = read_po(
        app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'definition_terms.po'
    )
    actual = read_po(app.outdir / 'definition_terms.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.id in actual_msg_ids


@sphinx_intl
@pytest.mark.sphinx('gettext', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_gettext')
def test_gettext_glossary_terms(app):
    app.build()
    # --- glossary terms: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/1090
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'glossary_terms.po')
    actual = read_po(app.outdir / 'glossary_terms.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.id in actual_msg_ids
    warnings = app.warning.getvalue().replace(os.sep, '/')
    assert 'term not in glossary' not in warnings


@sphinx_intl
@pytest.mark.sphinx('gettext', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_gettext')
def test_gettext_glossary_term_inconsistencies(app):
    app.build()
    # --- glossary term inconsistencies: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/1090
    expect = read_po(
        app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'glossary_terms_inconsistency.po'
    )
    actual = read_po(app.outdir / 'glossary_terms_inconsistency.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.id in actual_msg_ids


@sphinx_intl
@pytest.mark.sphinx('gettext', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_gettext')
def test_gettext_literalblock(app):
    app.build()
    # --- gettext builder always ignores ``only`` directive
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'literalblock.po')
    actual = read_po(app.outdir / 'literalblock.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        if len(expect_msg.id.splitlines()) == 1:
            # compare translations only labels
            assert expect_msg.id in actual_msg_ids
        else:
            pass  # skip code-blocks and literalblocks


@sphinx_intl
@pytest.mark.sphinx('gettext', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_gettext')
def test_gettext_buildr_ignores_only_directive(app):
    app.build()
    # --- gettext builder always ignores ``only`` directive
    expect = read_po(app.srcdir / _CATALOG_LOCALE / 'LC_MESSAGES' / 'only.po')
    actual = read_po(app.outdir / 'only.pot')
    actual_msg_ids = {msg.id for msg in actual if msg.id}  # pyright: ignore[reportUnhashable]
    for expect_msg in (msg for msg in expect if msg.id):
        assert expect_msg.id in actual_msg_ids


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl', copy_test_root=True)
def test_node_translated_attribute(app):
    app.build(filenames=[app.srcdir / 'translation_progress.txt'])

    doctree = app.env.get_doctree('translation_progress')

    translated_nodes = sum(1 for _ in doctree.findall(NodeMatcher(translated=True)))
    assert translated_nodes == 10 + 1  # 10 lines + title

    untranslated_nodes = sum(1 for _ in doctree.findall(NodeMatcher(translated=False)))
    assert untranslated_nodes == 2 + 2 + 1  # 2 lines + 2 lines + substitution reference


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl', copy_test_root=True)
def test_translation_progress_substitution(app):
    app.build(filenames=[app.srcdir / 'translation_progress.txt'])

    doctree = app.env.get_doctree('translation_progress')

    assert doctree[0][19][0] == '68.75%'  # 11 out of 16 lines are translated


@pytest.mark.sphinx(
    'html',
    testroot='intl',
    freshenv=True,
    confoverrides={
        'language': _CATALOG_LOCALE,
        'locale_dirs': ['.'],
        'gettext_compact': False,
        'translation_progress_classes': True,
    },
    copy_test_root=True,
)
def test_translation_progress_classes_true(app):
    app.build(filenames=[app.srcdir / 'translation_progress.txt'])

    doctree = app.env.get_doctree('translation_progress')

    # title
    assert 'translated' in doctree[0][0]['classes']

    # translated lines
    assert 'translated' in doctree[0][1]['classes']
    assert 'translated' in doctree[0][2]['classes']
    assert 'translated' in doctree[0][3]['classes']
    assert 'translated' in doctree[0][4]['classes']
    assert 'translated' in doctree[0][5]['classes']
    assert 'translated' in doctree[0][6]['classes']
    assert 'translated' in doctree[0][7]['classes']
    assert 'translated' in doctree[0][8]['classes']

    assert doctree[0][9]['classes'] == []  # comment node

    # idempotent
    assert 'translated' in doctree[0][10]['classes']
    assert 'translated' in doctree[0][11]['classes']

    assert doctree[0][12]['classes'] == []  # comment node

    # untranslated
    assert 'untranslated' in doctree[0][13]['classes']
    assert 'untranslated' in doctree[0][14]['classes']

    assert doctree[0][15]['classes'] == []  # comment node

    # missing
    assert 'untranslated' in doctree[0][16]['classes']
    assert 'untranslated' in doctree[0][17]['classes']

    assert doctree[0][18]['classes'] == []  # comment node

    # substitution reference
    assert 'untranslated' in doctree[0][19]['classes']

    assert len(doctree[0]) == 20


class _MockClock:
    """Object for mocking :func:`time.time_ns` (if needed).

    Use :meth:`sleep` to make this specific clock sleep for some time.
    """

    def time(self) -> int:
        """Nanosecond since 'fake' epoch."""
        raise NotImplementedError

    def sleep(self, ds: float) -> None:
        """Sleep *ds* seconds."""
        raise NotImplementedError


class _MockWindowsClock(_MockClock):
    """Object for mocking :func:`time.time_ns` on Windows platforms.

    The result is in 'nanoseconds' but with a microsecond resolution
    so that the division by 1_000 does not cause rounding issues.
    """

    def __init__(self) -> None:
        self.us: int = 0  # current microsecond 'tick'

    def time(self) -> int:
        ret = 1_000 * self.us
        self.us += 1
        return ret

    def sleep(self, ds: float) -> None:
        self.us += int(ds * 1e6)


class _MockUnixClock(_MockClock):
    """Object for mocking :func:`time.time_ns` on Unix platforms.

    Since nothing is needed for Unix platforms, this object acts as
    a proxy so that the API is the same as :class:`_MockWindowsClock`.
    """

    def time(self) -> int:
        return time.time_ns()

    def sleep(self, ds: float) -> None:
        time.sleep(ds)


@pytest.fixture
def mock_time_and_i18n() -> Iterator[tuple[pytest.MonkeyPatch, _MockClock]]:
    from sphinx.util.i18n import CatalogInfo

    # save the 'original' definition
    catalog_write_mo = CatalogInfo.write_mo

    def mock_write_mo(self, locale, use_fuzzy=False):
        catalog_write_mo(self, locale, use_fuzzy)
        # ensure that the .mo file being written has a correct fake timestamp
        _set_mtime_ns(self.mo_path, time.time_ns())

    # see: https://github.com/pytest-dev/pytest/issues/363
    with pytest.MonkeyPatch.context() as mock:
        clock: _MockClock
        if os.name == 'posix':
            clock = _MockUnixClock()
        else:
            # When using pytest.mark.parametrize() to emulate test repetition,
            # the teardown phase on Windows fails due to an error apparently in
            # the colorama.ansitowin32 module, so we forcibly disable colors.
            mock.setenv('NO_COLOR', '1')
            # apply the patch only for Windows
            clock = _MockWindowsClock()
            mock.setattr('time.time_ns', clock.time)
            # Use clock.sleep() to emulate time.sleep() but do not try
            # to mock the latter since this might break other libraries.
            mock.setattr('sphinx.util.i18n.CatalogInfo.write_mo', mock_write_mo)
        yield mock, clock


# use the same testroot as 'gettext' since the latter contains less PO files
@sphinx_intl
@pytest.mark.sphinx(
    'dummy',
    testroot='builder-gettext-dont-rebuild-mo',
    freshenv=True,
    copy_test_root=True,
)
def test_dummy_should_rebuild_mo(mock_time_and_i18n, make_app, app_params):
    mock, clock = mock_time_and_i18n
    assert os.name == 'posix' or clock.time() == 0

    args, kwargs = app_params
    app = make_app(*args, **kwargs)
    po_path, mo_path = _get_bom_intl_path(app)

    # creation time of the those files (order does not matter)
    bom_rst = app.srcdir / 'bom.rst'
    bom_rst_time = time.time_ns()

    index_rst = app.srcdir / 'index.rst'
    index_rst_time = time.time_ns()
    po_time = time.time_ns()

    # patch the 'creation time' of the source files
    assert _set_mtime_ns(po_path, po_time) == po_time
    assert _set_mtime_ns(bom_rst, bom_rst_time) == bom_rst_time
    assert _set_mtime_ns(index_rst, index_rst_time) == index_rst_time

    assert not mo_path.exists()
    # when writing mo files, the counter is updated by calling
    # patch_write_mo which is called to create .mo files (and
    # thus the timestamp of the files are not those given by
    # the OS but our fake ones)
    app.build()
    assert mo_path.exists()
    # Do a real sleep on POSIX, or simulate a sleep on Windows
    # to ensure that calls to time.time_ns() remain consistent.
    clock.sleep(0.1 if os.name == 'posix' else 1)

    # check that the source files were not modified
    assert bom_rst.stat().st_mtime_ns == bom_rst_time
    assert index_rst.stat().st_mtime_ns == index_rst_time
    # check that the 'bom' document is discovered after the .mo
    # file has been written on the disk (i.e., read_doc() is called
    # after the creation of the .mo files)
    assert app.env.all_docs['bom'] > mo_path.stat().st_mtime_ns // 1000

    # Since it is after the build, the number of documents to be updated is 0
    update_targets = _get_update_targets(app)
    assert update_targets[1] == set()
    # When rewriting the timestamp of mo file, the number of documents to be
    # updated will be changed.
    new_mo_time = time.time_ns()
    assert _set_mtime_ns(mo_path, new_mo_time) == new_mo_time
    update_targets = _get_update_targets(app)
    assert update_targets[1] == {'bom'}
    mock.undo()  # explicit call since it's not a context

    # remove all sources for the next test
    shutil.rmtree(app.srcdir, ignore_errors=True)
    time.sleep(0.1 if os.name == 'posix' else 0.5)  # real sleep


@sphinx_intl
@pytest.mark.sphinx(
    'gettext',
    testroot='builder-gettext-dont-rebuild-mo',
    freshenv=True,
    copy_test_root=True,
)
def test_gettext_dont_rebuild_mo(mock_time_and_i18n, app):
    mock, clock = mock_time_and_i18n
    assert os.name == 'posix' or clock.time() == 0

    assert app.srcdir.exists()

    # patch the 'creation time' of the source files
    bom_rst = app.srcdir / 'bom.rst'
    bom_rst_time = time.time_ns()
    assert _set_mtime_ns(bom_rst, bom_rst_time) == bom_rst_time

    index_rst = app.srcdir / 'index.rst'
    index_rst_time = time.time_ns()
    assert _set_mtime_ns(index_rst, index_rst_time) == index_rst_time

    # phase 1: create fake MO file in the src directory
    po_path, mo_path = _get_bom_intl_path(app)
    write_mo(mo_path, read_po(po_path))
    po_time = time.time_ns()
    assert _set_mtime_ns(po_path, po_time) == po_time

    # phase 2: build document with gettext builder.
    # The mo file in the srcdir directory is retained.
    app.build()
    # Do a real sleep on POSIX, or simulate a sleep on Windows
    # to ensure that calls to time.time_ns() remain consistent.
    clock.sleep(0.5 if os.name == 'posix' else 1)
    # Since it is after the build, the number of documents to be updated is 0
    update_targets = _get_update_targets(app)
    assert update_targets[1] == set()
    # Even if the timestamp of the mo file is updated, the number of documents
    # to be updated is 0. gettext builder does not rebuild because of mo update.
    new_mo_time = time.time_ns()
    assert _set_mtime_ns(mo_path, new_mo_time) == new_mo_time
    update_targets = _get_update_targets(app)
    assert update_targets[1] == set()
    mock.undo()  # remove the patch

    # remove all sources for the next test
    shutil.rmtree(app.srcdir, ignore_errors=True)
    time.sleep(0.1 if os.name == 'posix' else 0.5)  # real sleep


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_html_meta(app):
    app.build()
    # --- test for meta
    result = (app.outdir / 'index.html').read_text(encoding='utf8')
    expected_expr = (
        '<meta content="TESTDATA FOR I18N" name="description" translated="True" />'
    )
    assert expected_expr in result
    expected_expr = (
        '<meta content="I18N, SPHINX, MARKUP" name="keywords" translated="True" />'
    )
    assert expected_expr in result
    expected_expr = '<p class="caption" role="heading"><span class="caption-text">HIDDEN TOC</span></p>'
    assert expected_expr in result


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_html_footnotes(app):
    app.build()
    # --- test for
    # https://github.com/sphinx-doc/sphinx/issues/955
    # cant-build-html-with-footnotes-when-using
    # expect no error by build
    (app.outdir / 'footnote.html').read_text(encoding='utf8')


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_html_undefined_refs(app):
    app.build()
    # --- links to undefined reference
    result = (app.outdir / 'refs_inconsistency.html').read_text(encoding='utf8')

    expected_expr = (
        '<a class="reference external" href="https://www.example.com">reference</a>'
    )
    assert len(re.findall(expected_expr, result)) == 2

    expected_expr = '<a class="reference internal" href="#reference">reference</a>'
    assert len(re.findall(expected_expr, result)) == 0

    expected_expr = (
        '<a class="reference internal" '
        'href="#i18n-with-refs-inconsistency">I18N WITH '
        'REFS INCONSISTENCY</a>'
    )
    assert len(re.findall(expected_expr, result)) == 1


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_html_index_entries(app):
    app.build()
    # --- index entries: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/976
    result = (app.outdir / 'genindex.html').read_text(encoding='utf8')

    def wrap(tag, keyword):
        start_tag = '<%s[^>]*>' % tag
        end_tag = '</%s>' % tag
        return rf'{start_tag}\s*{keyword}\s*{end_tag}'

    def wrap_nest(parenttag, childtag, keyword):
        start_tag1 = '<%s[^>]*>' % parenttag
        start_tag2 = '<%s[^>]*>' % childtag
        return rf'{start_tag1}\s*{keyword}\s*{start_tag2}'

    expected_exprs = [
        wrap('h2', 'Symbols'),
        wrap('h2', 'C'),
        wrap('h2', 'E'),
        wrap('h2', 'F'),
        wrap('h2', 'M'),
        wrap('h2', 'N'),
        wrap('h2', 'R'),
        wrap('h2', 'S'),
        wrap('h2', 'T'),
        wrap('h2', 'V'),
        wrap('a', 'NEWSLETTER'),
        wrap('a', 'MAILING LIST'),
        wrap('a', 'RECIPIENTS LIST'),
        wrap('a', 'FIRST SECOND'),
        wrap('a', 'SECOND THIRD'),
        wrap('a', 'THIRD, FIRST'),
        wrap_nest('li', 'ul', 'ENTRY'),
        wrap_nest('li', 'ul', 'SEE'),
    ]
    for expr in expected_exprs:
        assert re.search(expr, result, re.MULTILINE), (
            f'{expr!r} did not match {result!r}'
        )


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_html_versionchanges(app):
    app.build()
    # --- versionchanges
    result = (app.outdir / 'versionchange.html').read_text(encoding='utf8')

    def get_content(result, name):
        matched = re.search(r'<div class="%s">\n*(.*?)</div>' % name, result, re.DOTALL)
        if matched:
            return matched.group(1)
        else:
            return ''

    expect1 = (
        """<p><span class="versionmodified deprecated">Deprecated since version 1.0: </span>"""
        """THIS IS THE <em>FIRST</em> PARAGRAPH OF DEPRECATED.</p>\n"""
        """<p>THIS IS THE <em>SECOND</em> PARAGRAPH OF DEPRECATED.</p>\n"""
    )
    matched_content = get_content(result, 'deprecated')
    assert matched_content == expect1

    expect2 = (
        """<p><span class="versionmodified added">Added in version 1.0: </span>"""
        """THIS IS THE <em>FIRST</em> PARAGRAPH OF VERSIONADDED.</p>\n"""
    )
    matched_content = get_content(result, 'versionadded')
    assert matched_content == expect2

    expect3 = (
        """<p><span class="versionmodified changed">Changed in version 1.0: </span>"""
        """THIS IS THE <em>FIRST</em> PARAGRAPH OF VERSIONCHANGED.</p>\n"""
    )
    matched_content = get_content(result, 'versionchanged')
    assert matched_content == expect3

    expect4 = (
        """<p><span class="versionmodified removed">Removed in version 1.0: </span>"""
        """THIS IS THE <em>FIRST</em> PARAGRAPH OF VERSIONREMOVED.</p>\n"""
    )
    matched_content = get_content(result, 'versionremoved')
    assert matched_content == expect4


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_html_docfields(app):
    app.build()
    # --- docfields
    # expect no error by build
    (app.outdir / 'docfields.html').read_text(encoding='utf8')


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_html_template(app):
    app.build()
    # --- gettext template
    result = (app.outdir / 'contents.html').read_text(encoding='utf8')
    assert 'WELCOME' in result
    assert 'SPHINX 2013.120' in result


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_html_rebuild_mo(app):
    app.build()
    # --- rebuild by .mo mtime
    app.build()
    _, updated, _ = _get_update_targets(app)
    assert updated == set()

    _, bom_file = _get_bom_intl_path(app)
    old_mtime = bom_file.stat().st_mtime
    new_mtime = old_mtime + (dt := 5)
    os.utime(bom_file, (new_mtime, new_mtime))
    assert old_mtime + dt == new_mtime, (old_mtime + dt, new_mtime)
    _, updated, _ = _get_update_targets(app)
    assert updated == {'bom'}


@sphinx_intl
@pytest.mark.sphinx('xml', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_xml_footnotes(app):
    app.build()
    # --- footnotes: regression test for fix
    # https://github.com/sphinx-doc/sphinx/issues/955,
    # https://github.com/sphinx-doc/sphinx/issues/1176
    et = etree_parse(app.outdir / 'footnote.xml')
    secs = et.findall('section')

    para0 = secs[0].findall('paragraph')
    assert_elem(
        para0[0],
        [
            'I18N WITH FOOTNOTE',
            'INCLUDE THIS CONTENTS',
            '2',
            '[ref]',
            '1',
            '100',
            '*',
            '. SECOND FOOTNOTE_REF',
            '100',
            '.',
        ],
        ['i18n-with-footnote', 'ref'],
    )

    # check node_id for footnote_references which refer same footnote
    # See: https://github.com/sphinx-doc/sphinx/issues/3002
    assert para0[0][4].text == para0[0][6].text == '100'
    assert para0[0][4].attrib['ids'] != para0[0][6].attrib['ids']

    footnote0 = secs[0].findall('footnote')
    assert_elem(footnote0[0], ['1', 'THIS IS A AUTO NUMBERED FOOTNOTE.'], None, ['1'])
    assert_elem(footnote0[1], ['100', 'THIS IS A NUMBERED FOOTNOTE.'], None, ['100'])
    assert_elem(
        footnote0[2], ['2', 'THIS IS A AUTO NUMBERED NAMED FOOTNOTE.'], None, ['named']
    )
    assert_elem(footnote0[3], ['*', 'THIS IS A AUTO SYMBOL FOOTNOTE.'], None, None)

    citation0 = secs[0].findall('citation')
    assert_elem(citation0[0], ['ref', 'THIS IS A NAMED FOOTNOTE.'], None, ['ref'])

    warnings = getwarning(app.warning)
    warning_expr = '.*/footnote.xml:\\d*: SEVERE: Duplicate ID: ".*".\n'
    assert not re.search(warning_expr, warnings), (
        f'{warning_expr!r} did match {warnings!r}'
    )


@sphinx_intl
@pytest.mark.sphinx('xml', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_xml_footnote_backlinks(app):
    app.build()
    # --- footnote backlinks: i18n test for
    # https://github.com/sphinx-doc/sphinx/issues/1058
    et = etree_parse(app.outdir / 'footnote.xml')
    secs = et.findall('section')

    para0 = secs[0].findall('paragraph')
    refs0 = para0[0].findall('footnote_reference')
    refid2id = {r.attrib.get('refid'): r.attrib.get('ids') for r in refs0}

    footnote0 = secs[0].findall('footnote')
    for footnote in footnote0:
        ids = footnote.attrib.get('ids')
        backrefs = footnote.attrib.get('backrefs').split()
        assert refid2id[ids] in backrefs


@sphinx_intl
@pytest.mark.sphinx('xml', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_xml_refs_in_python_domain(app):
    app.build()
    # --- refs in the Python domain
    et = etree_parse(app.outdir / 'refs_python_domain.xml')
    secs = et.findall('section')

    # regression test for fix
    # https://github.com/sphinx-doc/sphinx/issues/1363
    para0 = secs[0].findall('paragraph')
    assert_elem(
        para0[0],
        ['SEE THIS DECORATOR:', 'sensitive_variables()', '.'],
        ['sensitive.sensitive_variables'],
    )


@sphinx_intl
@pytest.mark.sphinx('xml', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_xml_keep_external_links(app):
    app.build()
    # --- keep external links: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/1044
    et = etree_parse(app.outdir / 'external_links.xml')
    secs = et.findall('section')

    para0 = secs[0].findall('paragraph')
    # external link check
    assert_elem(
        para0[0], ['EXTERNAL LINK TO', 'Python', '.'], ['https://python.org/index.html']
    )

    # internal link check
    assert_elem(
        para0[1], ['EXTERNAL LINKS', 'IS INTERNAL LINK.'], ['i18n-with-external-links']
    )

    # inline link check
    assert_elem(
        para0[2], ['INLINE LINK BY', 'THE SPHINX SITE', '.'], ['https://sphinx-doc.org']
    )

    # unnamed link check
    assert_elem(para0[3], ['UNNAMED', 'LINK', '.'], ['https://google.com'])

    # link target swapped translation
    para1 = secs[1].findall('paragraph')
    assert_elem(
        para1[0],
        ['LINK TO', 'external2', 'AND', 'external1', '.'],
        ['https://www.google.com/external2', 'https://www.google.com/external1'],
    )
    assert_elem(
        para1[1],
        ['LINK TO', 'THE PYTHON SITE', 'AND', 'THE SPHINX SITE', '.'],
        ['https://python.org', 'https://sphinx-doc.org'],
    )

    # multiple references in the same line
    para2 = secs[2].findall('paragraph')
    assert_elem(
        para2[0],
        [
            'LINK TO',
            'EXTERNAL LINKS',
            ',',
            'Python',
            ',',
            'THE SPHINX SITE',
            ',',
            'UNNAMED',
            'AND',
            'THE PYTHON SITE',
            '.',
        ],
        [
            'i18n-with-external-links',
            'https://python.org/index.html',
            'https://sphinx-doc.org',
            'https://google.com',
            'https://python.org',
        ],
    )


@sphinx_intl
@pytest.mark.sphinx('xml', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_xml_role_xref(app):
    app.build()
    # --- role xref: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/1090,
    # https://github.com/sphinx-doc/sphinx/issues/1193
    et = etree_parse(app.outdir / 'role_xref.xml')
    sec1, sec2 = et.findall('section')

    (para1,) = sec1.findall('paragraph')
    assert_elem(
        para1,
        [
            'LINK TO',
            "I18N ROCK'N ROLE XREF",
            ',',
            'CONTENTS',
            ',',
            'SOME NEW TERM',
            '.',
        ],
        ['i18n-role-xref', 'index', 'glossary_terms#term-Some-term'],
    )

    (sec1_1,) = sec1.findall('section')
    (title,) = sec1_1.findall('title')
    assert_elem(
        title,
        [
            'LINK TO',
            "I18N ROCK'N ROLE XREF",
            ',',
            'CONTENTS',
            ',',
            'SOME NEW TERM',
            '.',
        ],
        ['i18n-role-xref', 'index', 'glossary_terms#term-Some-term'],
    )

    para2 = sec2.findall('paragraph')
    assert_elem(
        para2[0],
        ['LINK TO', 'SOME OTHER NEW TERM', 'AND', 'SOME NEW TERM', '.'],
        ['glossary_terms#term-Some-other-term', 'glossary_terms#term-Some-term'],
    )
    assert_elem(
        para2[1],
        ['LINK TO', 'LABEL', 'AND', 'SAME TYPE LINKS', 'AND', 'SAME TYPE LINKS', '.'],
        ['i18n-role-xref', 'same-type-links', 'same-type-links'],
    )
    assert_elem(
        para2[2],
        ['LINK TO', 'I18N WITH GLOSSARY TERMS', 'AND', 'CONTENTS', '.'],
        ['glossary_terms', 'index'],
    )
    assert_elem(
        para2[3],
        ['LINK TO', '--module', 'AND', '-m', '.'],
        ['cmdoption-module', 'cmdoption-m'],
    )
    assert_elem(
        para2[4],
        ['LINK TO', 'env2', 'AND', 'env1', '.'],
        ['envvar-env2', 'envvar-env1'],
    )
    # TODO: how do I link token role to productionlist?
    assert_elem(para2[5], ['LINK TO', 'token2', 'AND', 'token1', '.'], [])
    assert_elem(
        para2[6],
        ['LINK TO', 'same-type-links', 'AND', 'i18n-role-xref', '.'],
        ['same-type-links', 'i18n-role-xref'],
    )


@sphinx_intl
@pytest.mark.sphinx('xml', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_xml_warnings(app):
    app.build()
    # warnings
    warnings = getwarning(app.warning)
    assert warnings.count('term not in glossary') == 1
    assert 'undefined label' not in warnings
    assert 'unknown document' not in warnings


@sphinx_intl
@pytest.mark.sphinx('xml', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_xml_label_targets(app):
    app.build()
    # --- label targets: regression test for
    # https://github.com/sphinx-doc/sphinx/issues/1193,
    # https://github.com/sphinx-doc/sphinx/issues/1265
    et = etree_parse(app.outdir / 'label_target.xml')
    secs = et.findall('section')

    para0 = secs[0].findall('paragraph')
    assert_elem(
        para0[0],
        [
            'X SECTION AND LABEL',
            'POINT TO',
            'implicit-target',
            'AND',
            'X SECTION AND LABEL',
            'POINT TO',
            'section-and-label',
            '.',
        ],
        ['implicit-target', 'section-and-label'],
    )

    para1 = secs[1].findall('paragraph')
    assert_elem(
        para1[0],
        [
            'X EXPLICIT-TARGET',
            'POINT TO',
            'explicit-target',
            'AND',
            'X EXPLICIT-TARGET',
            'POINT TO DUPLICATED ID LIKE',
            'id1',
            '.',
        ],
        ['explicit-target', 'id1'],
    )

    para2 = secs[2].findall('paragraph')
    assert_elem(
        para2[0],
        ['X IMPLICIT SECTION NAME', 'POINT TO', 'implicit-section-name', '.'],
        ['implicit-section-name'],
    )

    sec2 = secs[2].findall('section')

    para2_0 = sec2[0].findall('paragraph')
    assert_elem(para2_0[0], ['`X DUPLICATED SUB SECTION`_', 'IS BROKEN LINK.'], [])

    para3 = secs[3].findall('paragraph')
    assert_elem(
        para3[0],
        [
            'X',
            'bridge label',
            'IS NOT TRANSLATABLE BUT LINKED TO TRANSLATED SECTION TITLE.',
        ],
        ['label-bridged-target-section'],
    )
    assert_elem(
        para3[1],
        [
            'X',
            'bridge label',
            'POINT TO',
            'LABEL BRIDGED TARGET SECTION',
            'AND',
            'bridge label2',
            'POINT TO',
            'SECTION AND LABEL',
            '. THE SECOND APPEARED',
            'bridge label2',
            'POINT TO CORRECT TARGET.',
        ],
        ['label-bridged-target-section', 'section-and-label', 'section-and-label'],
    )


@sphinx_intl
@pytest.mark.sphinx('xml', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_xml_strange_markup(app):
    app.build()
    et = etree_parse(app.outdir / 'markup.xml')
    secs = et.findall('section')

    (subsec1,) = secs[0].findall('section')
    (title1,) = subsec1.findall('title')
    assert_elem(title1, ['1. TITLE STARTING WITH 1.'])


@sphinx_intl
@pytest.mark.sphinx('html', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_additional_targets_should_not_be_translated(app):
    if tuple(map(int, pygments.__version__.split('.')[:2])) >= (2, 19):
        sp = '<span class="w"> </span>'
    else:
        sp = ' '

    app.build()
    # [literalblock.txt]
    result = (app.outdir / 'literalblock.html').read_text(encoding='utf8')

    # title should be translated
    expected_expr = 'CODE-BLOCKS'
    assert_count(expected_expr, result, 2)

    # ruby code block should not be translated but be highlighted
    expected_expr = """<span class="s1">&#39;result&#39;</span>"""
    assert_count(expected_expr, result, 1)

    # C code block without lang should not be translated and *ruby* highlighted
    expected_expr = """<span class="c1">#include &lt;stdlib.h&gt;</span>"""
    assert_count(expected_expr, result, 1)

    # C code block with lang should not be translated but be *C* highlighted
    expected_expr = (
        """<span class="cp">#include</span>"""
        """<span class="w"> </span>"""
        """<span class="cpf">&lt;stdio.h&gt;</span>"""
    )
    assert_count(expected_expr, result, 1)

    # literal block in list item should not be translated
    expected_expr = (
        """<span class="n">literal</span>"""
        """<span class="o">-</span>"""
        """<span class="n">block</span>\n"""
        """<span class="k">in</span>"""
        """<span class="w"> </span>"""
        """<span class="n">list</span>"""
    )
    assert_count(expected_expr, result, 1)

    # doctest block should not be translated but be highlighted
    expected_expr = (
        """<span class="gp">&gt;&gt;&gt; </span>"""
        f"""<span class="kn">import</span>{sp}<span class="nn">sys</span>  """
        """<span class="c1"># sys importing</span>"""
    )
    assert_count(expected_expr, result, 1)

    # [raw.txt]

    result = (app.outdir / 'raw.html').read_text(encoding='utf8')

    # raw block should not be translated
    expected_expr = """<iframe src="https://sphinx-doc.org"></iframe></section>"""
    assert_count(expected_expr, result, 1)

    # [figure.txt]

    result = (app.outdir / 'figure.html').read_text(encoding='utf8')

    # src for image block should not be translated (alt is translated)
    expected_expr = """<img alt="I18N -&gt; IMG" src="_images/i18n.png" />"""
    assert_count(expected_expr, result, 1)

    # src for figure block should not be translated (alt is translated)
    expected_expr = """<img alt="IMG -&gt; I18N" src="_images/img.png" />"""
    assert_count(expected_expr, result, 1)


@sphinx_intl
@pytest.mark.sphinx(
    'html',
    testroot='intl',
    srcdir='test_additional_targets_should_be_translated',
    confoverrides={
        'language': _CATALOG_LOCALE,
        'locale_dirs': ['.'],
        'gettext_compact': False,
        'gettext_additional_targets': [
            'index',
            'literal-block',
            'doctest-block',
            'raw',
            'image',
        ],
    },
)
def test_additional_targets_should_be_translated(app):
    if tuple(map(int, pygments.__version__.split('.')[:2])) >= (2, 19):
        sp = '<span class="w"> </span>'
    else:
        sp = ' '

    app.build()
    # [literalblock.txt]
    result = (app.outdir / 'literalblock.html').read_text(encoding='utf8')

    # basic literal block should be translated
    expected_expr = (
        '<span class="n">THIS</span> <span class="n">IS</span>\n'
        '<span class="n">LITERAL</span> <span class="n">BLOCK</span>'
    )
    assert_count(expected_expr, result, 1)

    # literalinclude should be translated
    expected_expr = '<span class="s2">&quot;HTTPS://SPHINX-DOC.ORG&quot;</span>'
    assert_count(expected_expr, result, 1)

    # title should be translated
    expected_expr = 'CODE-BLOCKS'
    assert_count(expected_expr, result, 2)

    # ruby code block should be translated and be highlighted
    expected_expr = """<span class="s1">&#39;RESULT&#39;</span>"""
    assert_count(expected_expr, result, 1)

    # C code block without lang should be translated and *ruby* highlighted
    expected_expr = """<span class="c1">#include &lt;STDLIB.H&gt;</span>"""
    assert_count(expected_expr, result, 1)

    # C code block with lang should be translated and be *C* highlighted
    expected_expr = (
        """<span class="cp">#include</span>"""
        """<span class="w"> </span>"""
        """<span class="cpf">&lt;STDIO.H&gt;</span>"""
    )
    assert_count(expected_expr, result, 1)

    # literal block in list item should be translated
    expected_expr = (
        """<span class="no">LITERAL</span>"""
        """<span class="o">-</span>"""
        """<span class="no">BLOCK</span>\n"""
        """<span class="no">IN</span>"""
        """<span class="w"> </span>"""
        """<span class="no">LIST</span>"""
    )
    assert_count(expected_expr, result, 1)

    # doctest block should not be translated but be highlighted
    expected_expr = (
        """<span class="gp">&gt;&gt;&gt; </span>"""
        f"""<span class="kn">import</span>{sp}<span class="nn">sys</span>  """
        """<span class="c1"># SYS IMPORTING</span>"""
    )
    assert_count(expected_expr, result, 1)

    # 'noqa' comments should remain in literal blocks.
    assert_count('#noqa', result, 1)

    # [raw.txt]

    result = (app.outdir / 'raw.html').read_text(encoding='utf8')

    # raw block should be translated
    expected_expr = """<iframe src="HTTPS://SPHINX-DOC.ORG"></iframe></section>"""
    assert_count(expected_expr, result, 1)

    # [figure.txt]

    result = (app.outdir / 'figure.html').read_text(encoding='utf8')

    # alt and src for image block should be translated
    expected_expr = """<img alt="I18N -&gt; IMG" src="_images/img.png" />"""
    assert_count(expected_expr, result, 1)

    # alt and src for figure block should be translated
    expected_expr = """<img alt="IMG -&gt; I18N" src="_images/i18n.png" />"""
    assert_count(expected_expr, result, 1)


@pytest.mark.sphinx(
    'html',
    testroot='intl_substitution_definitions',
    confoverrides={
        'language': _CATALOG_LOCALE,
        'locale_dirs': ['.'],
        'gettext_compact': False,
        'gettext_additional_targets': [
            'index',
            'literal-block',
            'doctest-block',
            'raw',
            'image',
        ],
    },
    copy_test_root=True,
)
def test_additional_targets_should_be_translated_substitution_definitions(app):
    app.build(force_all=True)

    # [prolog_epilog_substitution.txt]

    result = (app.outdir / 'prolog_epilog_substitution.html').read_text(encoding='utf8')

    # alt and src for image block should be translated
    expected_expr = """<img alt="SUBST_PROLOG_2 TRANSLATED" src="_images/i18n.png" />"""
    assert_count(expected_expr, result, 1)

    # alt and src for image block should be translated
    expected_expr = """<img alt="SUBST_EPILOG_2 TRANSLATED" src="_images/img.png" />"""
    assert_count(expected_expr, result, 1)


@sphinx_intl
@pytest.mark.sphinx('text', testroot='intl')
@pytest.mark.test_params(shared_result='test_intl_basic')
def test_text_references(app):
    app.build(filenames=[app.srcdir / 'refs.txt'])

    warnings = app.warning.getvalue().replace(os.sep, '/')
    warning_expr = 'refs.txt:\\d+: ERROR: Unknown target name:'
    assert_count(warning_expr, warnings, 0)


@pytest.mark.sphinx(
    'text',
    testroot='intl_substitution_definitions',
    confoverrides={
        'language': _CATALOG_LOCALE,
        'locale_dirs': ['.'],
        'gettext_compact': False,
    },
    copy_test_root=True,
)
def test_text_prolog_epilog_substitution(app):
    app.build()

    result = (app.outdir / 'prolog_epilog_substitution.txt').read_text(encoding='utf8')

    assert result == (
        """\
1. I18N WITH PROLOGUE AND EPILOGUE SUBSTITUTIONS
************************************************

THIS IS CONTENT THAT CONTAINS prologue substitute text.

SUBSTITUTED IMAGE [image: SUBST_PROLOG_2 TRANSLATED][image] HERE.

THIS IS CONTENT THAT CONTAINS epilogue substitute text.

SUBSTITUTED IMAGE [image: SUBST_EPILOG_2 TRANSLATED][image] HERE.
"""
    )


@pytest.mark.usefixtures('_http_teapot')
@pytest.mark.sphinx(
    'dummy',
    testroot='images',
    srcdir='test_intl_images',
    confoverrides={'language': _CATALOG_LOCALE},
)
def test_image_glob_intl(app):
    app.build()

    # index.rst
    doctree = app.env.get_doctree('index')
    assert_node(
        doctree[0][1], nodes.image, uri='rimg.xx.png', candidates={'*': 'rimg.xx.png'}
    )

    assert isinstance(doctree[0][2], nodes.figure)
    assert_node(
        doctree[0][2][0],
        nodes.image,
        uri='rimg.xx.png',
        candidates={'*': 'rimg.xx.png'},
    )

    assert_node(
        doctree[0][3],
        nodes.image,
        uri='img.*',
        candidates={
            'application/pdf': 'img.pdf',
            'image/gif': 'img.gif',
            'image/png': 'img.png',
        },
    )

    assert isinstance(doctree[0][4], nodes.figure)
    assert_node(
        doctree[0][4][0],
        nodes.image,
        uri='img.*',
        candidates={
            'application/pdf': 'img.pdf',
            'image/gif': 'img.gif',
            'image/png': 'img.png',
        },
    )

    # subdir/index.rst
    doctree = app.env.get_doctree('subdir/index')
    assert_node(
        doctree[0][1],
        nodes.image,
        uri='subdir/rimg.xx.png',
        candidates={'*': 'subdir/rimg.xx.png'},
    )

    assert_node(
        doctree[0][2],
        nodes.image,
        uri='subdir/svgimg.*',
        candidates={
            'application/pdf': 'subdir/svgimg.pdf',
            'image/svg+xml': 'subdir/svgimg.xx.svg',
        },
    )

    assert isinstance(doctree[0][3], nodes.figure)
    assert_node(
        doctree[0][3][0],
        nodes.image,
        uri='subdir/svgimg.*',
        candidates={
            'application/pdf': 'subdir/svgimg.pdf',
            'image/svg+xml': 'subdir/svgimg.xx.svg',
        },
    )


@pytest.mark.usefixtures('_http_teapot')
@pytest.mark.sphinx(
    'dummy',
    testroot='images',
    srcdir='test_intl_images',
    confoverrides={
        'language': _CATALOG_LOCALE,
        'figure_language_filename': '{root}{ext}.{language}',
    },
)
def test_image_glob_intl_using_figure_language_filename(app):
    app.build()

    # index.rst
    doctree = app.env.get_doctree('index')
    assert_node(
        doctree[0][1], nodes.image, uri='rimg.png.xx', candidates={'*': 'rimg.png.xx'}
    )

    assert isinstance(doctree[0][2], nodes.figure)
    assert_node(
        doctree[0][2][0],
        nodes.image,
        uri='rimg.png.xx',
        candidates={'*': 'rimg.png.xx'},
    )

    assert_node(
        doctree[0][3],
        nodes.image,
        uri='img.*',
        candidates={
            'application/pdf': 'img.pdf',
            'image/gif': 'img.gif',
            'image/png': 'img.png',
        },
    )

    assert isinstance(doctree[0][4], nodes.figure)
    assert_node(
        doctree[0][4][0],
        nodes.image,
        uri='img.*',
        candidates={
            'application/pdf': 'img.pdf',
            'image/gif': 'img.gif',
            'image/png': 'img.png',
        },
    )

    # subdir/index.rst
    doctree = app.env.get_doctree('subdir/index')
    assert_node(
        doctree[0][1],
        nodes.image,
        uri='subdir/rimg.png',
        candidates={'*': 'subdir/rimg.png'},
    )

    assert_node(
        doctree[0][2],
        nodes.image,
        uri='subdir/svgimg.*',
        candidates={
            'application/pdf': 'subdir/svgimg.pdf',
            'image/svg+xml': 'subdir/svgimg.svg',
        },
    )

    assert isinstance(doctree[0][3], nodes.figure)
    assert_node(
        doctree[0][3][0],
        nodes.image,
        uri='subdir/svgimg.*',
        candidates={
            'application/pdf': 'subdir/svgimg.pdf',
            'image/svg+xml': 'subdir/svgimg.svg',
        },
    )


def getwarning(warnings: StringIO) -> str:
    return strip_escape_sequences(warnings.getvalue().replace(os.sep, '/'))


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    srcdir='gettext_allow_fuzzy_translations',
    confoverrides={
        'language': 'de',
        'gettext_allow_fuzzy_translations': True,
    },
)
def test_gettext_allow_fuzzy_translations(app):
    locale_dir = app.srcdir / 'locales' / 'de' / 'LC_MESSAGES'
    locale_dir.mkdir(parents=True, exist_ok=True)
    with (locale_dir / 'index.po').open('wb') as f:
        catalog = Catalog()
        catalog.add('features', 'FEATURES', flags=('fuzzy',))
        pofile.write_po(f, catalog)

    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert 'FEATURES' in content


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    srcdir='gettext_disallow_fuzzy_translations',
    confoverrides={
        'language': 'de',
        'gettext_allow_fuzzy_translations': False,
    },
)
def test_gettext_disallow_fuzzy_translations(app):
    locale_dir = app.srcdir / 'locales' / 'de' / 'LC_MESSAGES'
    locale_dir.mkdir(parents=True, exist_ok=True)
    with (locale_dir / 'index.po').open('wb') as f:
        catalog = Catalog()
        catalog.add('features', 'FEATURES', flags=('fuzzy',))
        pofile.write_po(f, catalog)

    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    assert 'FEATURES' not in content


@pytest.mark.sphinx(
    'html',
    testroot='basic',
    confoverrides={'language': 'de', 'html_sidebars': {'**': ['searchbox.html']}},
    copy_test_root=True,
)
def test_customize_system_message(make_app, app_params):
    try:
        # clear translators cache
        locale.translators.clear()

        # prepare message catalog (.po)
        locale_dir = app_params.kwargs['srcdir'] / 'locales' / 'de' / 'LC_MESSAGES'
        locale_dir.mkdir(parents=True, exist_ok=True)
        with (locale_dir / 'sphinx.po').open('wb') as f:
            catalog = Catalog()
            catalog.add('Quick search', 'QUICK SEARCH')
            pofile.write_po(f, catalog)

        # construct application and convert po file to .mo
        args, kwargs = app_params
        app = make_app(*args, **kwargs)
        assert (locale_dir / 'sphinx.mo').exists()
        assert app.translator.gettext('Quick search') == 'QUICK SEARCH'

        app.build()
        content = (app.outdir / 'index.html').read_text(encoding='utf8')
        assert 'QUICK SEARCH' in content
    finally:
        locale.translators.clear()


@pytest.mark.sphinx(
    'html',
    testroot='intl',
    confoverrides={'today_fmt': '%Y-%m-%d'},
)
def test_customize_today_date_format(app, monkeypatch):
    with monkeypatch.context() as m:
        m.setenv('SOURCE_DATE_EPOCH', '1439131307')
        app.build()
        content = (app.outdir / 'refs.html').read_text(encoding='utf8')

    assert '2015-08-09' in content



"""Tests the JavaScript Domain"""

from __future__ import annotations

from unittest.mock import Mock

import docutils.utils
import pytest
from docutils import nodes

from sphinx import addnodes
from sphinx.addnodes import (
    desc,
    desc_annotation,
    desc_content,
    desc_name,
    desc_parameter,
    desc_parameterlist,
    desc_sig_keyword,
    desc_sig_name,
    desc_sig_space,
    desc_signature,
)
from sphinx.domains.javascript import JavaScriptDomain
from sphinx.testing import restructuredtext
from sphinx.testing.util import assert_node
from sphinx.writers.text import STDINDENT


@pytest.mark.sphinx('dummy', testroot='domain-js')
def test_domain_js_xrefs(app):
    """Domain objects have correct prefixes when looking up xrefs"""
    app.build(force_all=True)

    def assert_refnode(node, mod_name, prefix, target, reftype=None, domain='js'):
        attributes = {
            'refdomain': domain,
            'reftarget': target,
        }
        if reftype is not None:
            attributes['reftype'] = reftype
        if mod_name is not False:
            attributes['js:module'] = mod_name
        if prefix is not False:
            attributes['js:object'] = prefix
        assert_node(node, **attributes)

    doctree = app.env.get_doctree('roles')
    refnodes = list(doctree.findall(addnodes.pending_xref))
    assert_refnode(refnodes[0], None, None, 'TopLevel', 'class')
    assert_refnode(refnodes[1], None, None, 'top_level', 'func')
    assert_refnode(refnodes[2], None, 'NestedParentA', 'child_1', 'func')
    assert_refnode(
        refnodes[3], None, 'NestedParentA', 'NestedChildA.subchild_2', 'func'
    )
    assert_refnode(refnodes[4], None, 'NestedParentA', 'child_2', 'func')
    assert_refnode(refnodes[5], False, 'NestedParentA', 'any_child', domain='')
    assert_refnode(refnodes[6], None, 'NestedParentA', 'NestedChildA', 'class')
    assert_refnode(
        refnodes[7], None, 'NestedParentA.NestedChildA', 'subchild_2', 'func'
    )
    assert_refnode(
        refnodes[8], None, 'NestedParentA.NestedChildA', 'NestedParentA.child_1', 'func'
    )
    assert_refnode(
        refnodes[9], None, 'NestedParentA', 'NestedChildA.subchild_1', 'func'
    )
    assert_refnode(refnodes[10], None, 'NestedParentB', 'child_1', 'func')
    assert_refnode(refnodes[11], None, 'NestedParentB', 'NestedParentB', 'class')
    assert_refnode(refnodes[12], None, None, 'NestedParentA.NestedChildA', 'class')
    assert len(refnodes) == 13

    doctree = app.env.get_doctree('module')
    refnodes = list(doctree.findall(addnodes.pending_xref))
    assert_refnode(refnodes[0], 'module_a.submodule', None, 'ModTopLevel', 'class')
    assert_refnode(
        refnodes[1], 'module_a.submodule', 'ModTopLevel', 'mod_child_1', 'meth'
    )
    assert_refnode(
        refnodes[2],
        'module_a.submodule',
        'ModTopLevel',
        'ModTopLevel.mod_child_1',
        'meth',
    )
    assert_refnode(
        refnodes[3], 'module_a.submodule', 'ModTopLevel', 'mod_child_2', 'meth'
    )
    assert_refnode(
        refnodes[4],
        'module_a.submodule',
        'ModTopLevel',
        'module_a.submodule.ModTopLevel.mod_child_1',
        'meth',
    )
    assert_refnode(refnodes[5], 'module_b.submodule', None, 'ModTopLevel', 'class')
    assert_refnode(
        refnodes[6], 'module_b.submodule', 'ModTopLevel', 'module_a.submodule', 'mod'
    )
    assert len(refnodes) == 7


@pytest.mark.sphinx('dummy', testroot='domain-js')
def test_domain_js_objects(app):
    app.build(force_all=True)

    modules = app.env.domains.javascript_domain.data['modules']
    objects = app.env.domains.javascript_domain.data['objects']

    assert 'module_a.submodule' in modules
    assert 'module_a.submodule' in objects
    assert 'module_b.submodule' in modules
    assert 'module_b.submodule' in objects

    assert objects['module_a.submodule.ModTopLevel'][2] == 'class'
    assert objects['module_a.submodule.ModTopLevel.mod_child_1'][2] == 'method'
    assert objects['module_a.submodule.ModTopLevel.mod_child_2'][2] == 'method'
    assert objects['module_b.submodule.ModTopLevel'][2] == 'class'

    assert objects['TopLevel'][2] == 'class'
    assert objects['top_level'][2] == 'function'
    assert objects['NestedParentA'][2] == 'class'
    assert objects['NestedParentA.child_1'][2] == 'function'
    assert objects['NestedParentA.any_child'][2] == 'function'
    assert objects['NestedParentA.NestedChildA'][2] == 'class'
    assert objects['NestedParentA.NestedChildA.subchild_1'][2] == 'function'
    assert objects['NestedParentA.NestedChildA.subchild_2'][2] == 'function'
    assert objects['NestedParentA.child_2'][2] == 'function'
    assert objects['NestedParentB'][2] == 'class'
    assert objects['NestedParentB.child_1'][2] == 'function'


@pytest.mark.sphinx('dummy', testroot='domain-js')
def test_domain_js_find_obj(app):
    def find_obj(mod_name, prefix, obj_name, obj_type, searchmode=0):
        return app.env.domains.javascript_domain.find_obj(
            app.env, mod_name, prefix, obj_name, obj_type, searchmode
        )

    app.build(force_all=True)

    assert find_obj(None, None, 'NONEXISTANT', 'class') == (None, None)
    assert find_obj(None, None, 'NestedParentA', 'class') == (
        'NestedParentA',
        ('roles', 'NestedParentA', 'class'),
    )
    assert find_obj(None, None, 'NestedParentA.NestedChildA', 'class') == (
        'NestedParentA.NestedChildA',
        ('roles', 'NestedParentA.NestedChildA', 'class'),
    )
    assert find_obj(None, 'NestedParentA', 'NestedChildA', 'class') == (
        'NestedParentA.NestedChildA',
        ('roles', 'NestedParentA.NestedChildA', 'class'),
    )
    assert find_obj(None, None, 'NestedParentA.NestedChildA.subchild_1', 'func') == (
        'NestedParentA.NestedChildA.subchild_1',
        ('roles', 'NestedParentA.NestedChildA.subchild_1', 'function'),
    )
    assert find_obj(None, 'NestedParentA', 'NestedChildA.subchild_1', 'func') == (
        'NestedParentA.NestedChildA.subchild_1',
        ('roles', 'NestedParentA.NestedChildA.subchild_1', 'function'),
    )
    assert find_obj(None, 'NestedParentA.NestedChildA', 'subchild_1', 'func') == (
        'NestedParentA.NestedChildA.subchild_1',
        ('roles', 'NestedParentA.NestedChildA.subchild_1', 'function'),
    )
    assert find_obj('module_a.submodule', 'ModTopLevel', 'mod_child_2', 'meth') == (
        'module_a.submodule.ModTopLevel.mod_child_2',
        ('module', 'module_a.submodule.ModTopLevel.mod_child_2', 'method'),
    )
    assert find_obj(
        'module_b.submodule', 'ModTopLevel', 'module_a.submodule', 'mod'
    ) == ('module_a.submodule', ('module', 'module-module_a.submodule', 'module'))


def test_get_full_qualified_name() -> None:
    env = Mock(domaindata={})
    domain = JavaScriptDomain(env)

    # non-js references
    node = nodes.reference()
    assert domain.get_full_qualified_name(node) is None

    # simple reference
    node = nodes.reference(reftarget='func')
    assert domain.get_full_qualified_name(node) == 'func'

    # with js:module context
    kwargs = {'js:module': 'module1'}
    node = nodes.reference(reftarget='func', **kwargs)
    assert domain.get_full_qualified_name(node) == 'module1.func'

    # with js:object context
    kwargs = {'js:object': 'Class'}
    node = nodes.reference(reftarget='func', **kwargs)
    assert domain.get_full_qualified_name(node) == 'Class.func'

    # with both js:module and js:object context
    kwargs = {'js:module': 'module1', 'js:object': 'Class'}
    node = nodes.reference(reftarget='func', **kwargs)
    assert domain.get_full_qualified_name(node) == 'module1.Class.func'


@pytest.mark.sphinx('html', testroot='_blank')
def test_js_module(app):
    text = '.. js:module:: sphinx'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, nodes.target))
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('single', 'sphinx (module)', 'module-sphinx', '', None)],
    )
    assert_node(doctree[1], nodes.target, ids=['module-sphinx'])


@pytest.mark.sphinx('html', testroot='_blank')
def test_js_function(app):
    text = '.. js:function:: sum(a, b)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        ([desc_name, ([desc_sig_name, 'sum'])], desc_parameterlist),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[1][0][1],
        [
            desc_parameterlist,
            (
                [desc_parameter, ([desc_sig_name, 'a'])],
                [desc_parameter, ([desc_sig_name, 'b'])],
            ),
        ],
    )
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('single', 'sum() (built-in function)', 'sum', '', None)],
    )
    assert_node(
        doctree[1], addnodes.desc, domain='js', objtype='function', no_index=False
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_js_class(app):
    text = '.. js:class:: Application'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [
                                desc_annotation,
                                ([desc_sig_keyword, 'class'], desc_sig_space),
                            ],
                            [desc_name, ([desc_sig_name, 'Application'])],
                            [desc_parameterlist, ()],
                        ),
                    ],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('single', 'Application() (class)', 'Application', '', None)],
    )
    assert_node(doctree[1], addnodes.desc, domain='js', objtype='class', no_index=False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_js_data(app):
    text = '.. js:data:: name'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [desc_signature, ([desc_name, ([desc_sig_name, 'name'])])],
                    [desc_content, ()],
                ),
            ],
        ),
    )
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('single', 'name (global variable or constant)', 'name', '', None)],
    )
    assert_node(doctree[1], addnodes.desc, domain='js', objtype='data', no_index=False)


@pytest.mark.sphinx('html', testroot='_blank')
def test_no_index_entry(app):
    text = '.. js:function:: f()\n.. js:function:: g()\n   :no-index-entry:\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, desc, addnodes.index, desc))
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('single', 'f() (built-in function)', 'f', '', None)],
    )
    assert_node(doctree[2], addnodes.index, entries=[])

    text = '.. js:class:: f\n.. js:class:: g\n   :no-index-entry:\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, desc, addnodes.index, desc))
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('single', 'f() (class)', 'f', '', None)],
    )
    assert_node(doctree[2], addnodes.index, entries=[])

    text = '.. js:module:: f\n.. js:module:: g\n   :no-index-entry:\n'
    doctree = restructuredtext.parse(app, text)
    assert_node(doctree, (addnodes.index, nodes.target, nodes.target))
    assert_node(
        doctree[0],
        addnodes.index,
        entries=[('single', 'f (module)', 'module-f', '', None)],
    )


@pytest.mark.sphinx('html', testroot='_blank')
def test_module_content_line_number(app):
    text = '.. js:module:: foo\n\n   Some link here: :ref:`abc`\n'
    doc = restructuredtext.parse(app, text)
    xrefs = list(doc.findall(condition=addnodes.pending_xref))
    assert len(xrefs) == 1
    source, line = docutils.utils.get_source_line(xrefs[0])
    assert 'index.rst' in source
    assert line == 3


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'javascript_maximum_signature_line_length': len('hello(name)'),
    },
)
def test_jsfunction_signature_with_javascript_maximum_signature_line_length_equal(app):
    text = '.. js:function:: hello(name)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, ([desc_sig_name, 'hello'])],
                            desc_parameterlist,
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        desc,
        desctype='function',
        domain='js',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [desc_parameterlist, desc_parameter, ([desc_sig_name, 'name'])],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'javascript_maximum_signature_line_length': len('hello(name)'),
    },
)
def test_jsfunction_signature_with_javascript_maximum_signature_line_length_force_single(
    app,
):
    text = '.. js:function:: hello(names)\n   :single-line-parameter-list:'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, ([desc_sig_name, 'hello'])],
                            desc_parameterlist,
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        desc,
        desctype='function',
        domain='js',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [desc_parameterlist, desc_parameter, ([desc_sig_name, 'names'])],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'javascript_maximum_signature_line_length': len('hello(name)'),
    },
)
def test_jsfunction_signature_with_javascript_maximum_signature_line_length_break(app):
    text = '.. js:function:: hello(names)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, ([desc_sig_name, 'hello'])],
                            desc_parameterlist,
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        desc,
        desctype='function',
        domain='js',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [desc_parameterlist, desc_parameter, ([desc_sig_name, 'names'])],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=True)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'maximum_signature_line_length': len('hello(name)'),
    },
)
def test_jsfunction_signature_with_maximum_signature_line_length_equal(app):
    text = '.. js:function:: hello(name)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, ([desc_sig_name, 'hello'])],
                            desc_parameterlist,
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        desc,
        desctype='function',
        domain='js',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [desc_parameterlist, desc_parameter, ([desc_sig_name, 'name'])],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'maximum_signature_line_length': len('hello(name)'),
    },
)
def test_jsfunction_signature_with_maximum_signature_line_length_force_single(app):
    text = '.. js:function:: hello(names)\n   :single-line-parameter-list:'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, ([desc_sig_name, 'hello'])],
                            desc_parameterlist,
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        desc,
        desctype='function',
        domain='js',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [desc_parameterlist, desc_parameter, ([desc_sig_name, 'names'])],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'maximum_signature_line_length': len('hello(name)'),
    },
)
def test_jsfunction_signature_with_maximum_signature_line_length_break(app):
    text = '.. js:function:: hello(names)'
    doctree = restructuredtext.parse(app, text)
    assert_node(
        doctree,
        (
            addnodes.index,
            [
                desc,
                (
                    [
                        desc_signature,
                        (
                            [desc_name, ([desc_sig_name, 'hello'])],
                            desc_parameterlist,
                        ),
                    ],
                    desc_content,
                ),
            ],
        ),
    )
    assert_node(
        doctree[1],
        desc,
        desctype='function',
        domain='js',
        objtype='function',
        no_index=False,
    )
    assert_node(
        doctree[1][0][1],
        [desc_parameterlist, desc_parameter, ([desc_sig_name, 'names'])],
    )
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=True)


@pytest.mark.sphinx(
    'html',
    testroot='root',
    confoverrides={
        'javascript_maximum_signature_line_length': len('hello(name)'),
        'maximum_signature_line_length': 1,
    },
)
def test_javascript_maximum_signature_line_length_overrides_global(app):
    text = '.. js:function:: hello(name)'
    doctree = restructuredtext.parse(app, text)
    expected_doctree = (
        addnodes.index,
        [
            desc,
            (
                [
                    desc_signature,
                    ([desc_name, ([desc_sig_name, 'hello'])], desc_parameterlist),
                ],
                desc_content,
            ),
        ],
    )
    assert_node(doctree, expected_doctree)
    assert_node(
        doctree[1],
        desc,
        desctype='function',
        domain='js',
        objtype='function',
        no_index=False,
    )
    expected_sig = [desc_parameterlist, desc_parameter, [desc_sig_name, 'name']]
    assert_node(doctree[1][0][1], expected_sig)
    assert_node(doctree[1][0][1], desc_parameterlist, multi_line_parameter_list=False)


@pytest.mark.sphinx(
    'html',
    testroot='domain-js-javascript_maximum_signature_line_length',
)
def test_domain_js_javascript_maximum_signature_line_length_in_html(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    expected_parameter_list_hello = """\

<dl>
<dd>\
<em class="sig-param">\
<span class="n"><span class="pre">name</span></span>\
</em>,\
</dd>
</dl>

<span class="sig-paren">)</span>\
<a class="headerlink" href="#hello" title="Link to this definition">¶</a>\
</dt>\
"""
    assert expected_parameter_list_hello in content

    param_line_fmt = '<dd>{}</dd>\n'
    param_name_fmt = (
        '<em class="sig-param"><span class="n"><span class="pre">{}</span></span></em>'
    )
    optional_fmt = '<span class="optional">{}</span>'

    expected_a = param_line_fmt.format(
        optional_fmt.format('[')
        + param_name_fmt.format('a')
        + ','
        + optional_fmt.format('['),
    )
    assert expected_a in content

    expected_b = param_line_fmt.format(
        param_name_fmt.format('b')
        + ','
        + optional_fmt.format(']')
        + optional_fmt.format(']'),
    )
    assert expected_b in content

    expected_c = param_line_fmt.format(param_name_fmt.format('c') + ',')
    assert expected_c in content

    expected_d = param_line_fmt.format(
        param_name_fmt.format('d') + optional_fmt.format('[') + ','
    )
    assert expected_d in content

    expected_e = param_line_fmt.format(param_name_fmt.format('e') + ',')
    assert expected_e in content

    expected_f = param_line_fmt.format(
        param_name_fmt.format('f') + ',' + optional_fmt.format(']')
    )
    assert expected_f in content

    expected_parameter_list_foo = """\

<dl>
{}{}{}{}{}{}</dl>

<span class="sig-paren">)</span>\
<a class="headerlink" href="#foo" title="Link to this definition">¶</a>\
</dt>\
""".format(expected_a, expected_b, expected_c, expected_d, expected_e, expected_f)
    assert expected_parameter_list_foo in content


@pytest.mark.sphinx(
    'text',
    testroot='domain-js-javascript_maximum_signature_line_length',
)
def test_domain_js_javascript_maximum_signature_line_length_in_text(app):
    app.build()
    content = (app.outdir / 'index.txt').read_text(encoding='utf8')
    param_line_fmt = STDINDENT * ' ' + '{}\n'

    expected_parameter_list_hello = '(\n{})'.format(param_line_fmt.format('name,'))

    assert expected_parameter_list_hello in content

    expected_a = param_line_fmt.format('[a,[')
    assert expected_a in content

    expected_b = param_line_fmt.format('b,]]')
    assert expected_b in content

    expected_c = param_line_fmt.format('c,')
    assert expected_c in content

    expected_d = param_line_fmt.format('d[,')
    assert expected_d in content

    expected_e = param_line_fmt.format('e,')
    assert expected_e in content

    expected_f = param_line_fmt.format('f,]')
    assert expected_f in content

    expected_parameter_list_foo = '(\n{}{}{}{}{}{})'.format(
        expected_a,
        expected_b,
        expected_c,
        expected_d,
        expected_e,
        expected_f,
    )
    assert expected_parameter_list_foo in content


@pytest.mark.sphinx(
    'html',
    testroot='domain-js-javascript_maximum_signature_line_length',
    confoverrides={'javascript_trailing_comma_in_multi_line_signatures': False},
)
def test_domain_js_javascript_trailing_comma_in_multi_line_signatures_in_html(app):
    app.build()
    content = (app.outdir / 'index.html').read_text(encoding='utf8')
    expected_parameter_list_hello = """\

<dl>
<dd>\
<em class="sig-param">\
<span class="n"><span class="pre">name</span></span>\
</em>\
</dd>
</dl>

<span class="sig-paren">)</span>\
<a class="headerlink" href="#hello" title="Link to this definition">¶</a>\
</dt>\
"""
    assert expected_parameter_list_hello in content

    param_line_fmt = '<dd>{}</dd>\n'
    param_name_fmt = (
        '<em class="sig-param"><span class="n"><span class="pre">{}</span></span></em>'
    )
    optional_fmt = '<span class="optional">{}</span>'

    expected_a = param_line_fmt.format(
        optional_fmt.format('[')
        + param_name_fmt.format('a')
        + ','
        + optional_fmt.format('['),
    )
    assert expected_a in content

    expected_b = param_line_fmt.format(
        param_name_fmt.format('b')
        + ','
        + optional_fmt.format(']')
        + optional_fmt.format(']'),
    )
    assert expected_b in content

    expected_c = param_line_fmt.format(param_name_fmt.format('c') + ',')
    assert expected_c in content

    expected_d = param_line_fmt.format(
        param_name_fmt.format('d') + optional_fmt.format('[') + ','
    )
    assert expected_d in content

    expected_e = param_line_fmt.format(param_name_fmt.format('e') + ',')
    assert expected_e in content

    expected_f = param_line_fmt.format(
        param_name_fmt.format('f') + optional_fmt.format(']')
    )
    assert expected_f in content

    expected_parameter_list_foo = """\

<dl>
{}{}{}{}{}{}</dl>

<span class="sig-paren">)</span>\
<a class="headerlink" href="#foo" title="Link to this definition">¶</a>\
</dt>\
""".format(expected_a, expected_b, expected_c, expected_d, expected_e, expected_f)
    assert expected_parameter_list_foo in content


@pytest.mark.sphinx(
    'text',
    testroot='domain-js-javascript_maximum_signature_line_length',
    freshenv=True,
    confoverrides={'javascript_trailing_comma_in_multi_line_signatures': False},
)
def test_domain_js_javascript_trailing_comma_in_multi_line_signatures_in_text(app):
    app.build()
    content = (app.outdir / 'index.txt').read_text(encoding='utf8')
    param_line_fmt = STDINDENT * ' ' + '{}\n'

    expected_parameter_list_hello = '(\n{})'.format(param_line_fmt.format('name'))

    assert expected_parameter_list_hello in content

    expected_a = param_line_fmt.format('[a,[')
    assert expected_a in content

    expected_b = param_line_fmt.format('b,]]')
    assert expected_b in content

    expected_c = param_line_fmt.format('c,')
    assert expected_c in content

    expected_d = param_line_fmt.format('d[,')
    assert expected_d in content

    expected_e = param_line_fmt.format('e,')
    assert expected_e in content

    expected_f = param_line_fmt.format('f]')
    assert expected_f in content

    expected_parameter_list_foo = '(\n{}{}{}{}{}{})'.format(
        expected_a,
        expected_b,
        expected_c,
        expected_d,
        expected_e,
        expected_f,
    )
    assert expected_parameter_list_foo in content



from __future__ import annotations

import sys
from pathlib import Path

from sphinx.ext.autodoc.importer import import_module


def test_import_native_module_stubs(rootdir: Path) -> None:
    fish_licence_root = rootdir / 'test-ext-apidoc-duplicates'

    sys_path = list(sys.path)
    sys.path.insert(0, str(fish_licence_root))
    halibut = import_module('fish_licence.halibut')
    sys.path[:] = sys_path

    assert halibut.__file__.endswith('halibut.pyi')
    assert halibut.__spec__.origin.endswith('halibut.pyi')

    halibut_path = Path(halibut.__file__).resolve()
    assert halibut_path.is_file()
    assert halibut_path == fish_licence_root / 'fish_licence' / 'halibut.pyi'



"""Test the sphinx.apidoc module."""

from __future__ import annotations

from collections import namedtuple
from typing import TYPE_CHECKING

import pytest

import sphinx.ext.apidoc._generate
from sphinx.ext.apidoc._cli import main as apidoc_main

if TYPE_CHECKING:
    from pathlib import Path

    from sphinx.testing.util import SphinxTestApp

_apidoc = namedtuple('_apidoc', 'coderoot,outdir')  # NoQA: PYI024


@pytest.fixture
def apidoc(rootdir, tmp_path, apidoc_params):
    _, kwargs = apidoc_params
    coderoot = rootdir / kwargs.get('coderoot', 'test-root')
    outdir = tmp_path / 'out'
    excludes = [str(coderoot / e) for e in kwargs.get('excludes', [])]
    args = [
        '-o',
        str(outdir),
        '-F',
        str(coderoot),
        *excludes,
        *kwargs.get('options', []),
    ]
    apidoc_main(args)
    return _apidoc(coderoot, outdir)


@pytest.fixture
def apidoc_params(request):
    pargs = {}
    kwargs = {}

    for info in reversed(list(request.node.iter_markers('apidoc'))):
        pargs |= dict(enumerate(info.args))
        kwargs.update(info.kwargs)

    args = [pargs[i] for i in sorted(pargs.keys())]
    return args, kwargs


@pytest.mark.apidoc(coderoot='test-root')
def test_simple(make_app, apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'index.rst').is_file()

    app = make_app('text', srcdir=outdir)
    app.build()
    print(app._status.getvalue())
    print(app._warning.getvalue())


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-custom-templates',
    options=[
        '--separate',
        '--templatedir=tests/roots/test-ext-apidoc-custom-templates/_templates',
    ],
)
def test_custom_templates(make_app, apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'index.rst').is_file()

    template_dir = apidoc.coderoot / '_templates'
    assert sorted(template_dir.iterdir()) == [
        template_dir / 'module.rst.jinja',
        template_dir / 'module.rst_t',
        template_dir / 'package.rst_t',
    ]

    app = make_app('text', srcdir=outdir)
    app.build()

    builddir = outdir / '_build' / 'text'

    # Assert that the legacy filename is discovered
    with open(builddir / 'mypackage.txt', encoding='utf-8') as f:
        txt = f.read()
    assert 'The legacy package template was found!' in txt

    # Assert that the new filename is preferred
    with open(builddir / 'mypackage.mymodule.txt', encoding='utf-8') as f:
        txt = f.read()
    assert 'The Jinja module template was found!' in txt


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-pep420/a',
    options=['--implicit-namespaces'],
)
def test_pep_0420_enabled(make_app, apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'a.b.c.rst').is_file()
    assert (outdir / 'a.b.e.rst').is_file()
    assert (outdir / 'a.b.x.rst').is_file()

    with open(outdir / 'a.b.c.rst', encoding='utf-8') as f:
        rst = f.read()
    assert 'automodule:: a.b.c.d\n' in rst
    assert 'automodule:: a.b.c\n' in rst

    with open(outdir / 'a.b.e.rst', encoding='utf-8') as f:
        rst = f.read()
    assert 'automodule:: a.b.e.f\n' in rst

    with open(outdir / 'a.b.x.rst', encoding='utf-8') as f:
        rst = f.read()
    assert 'automodule:: a.b.x.y\n' in rst
    assert 'automodule:: a.b.x\n' not in rst

    app = make_app('text', srcdir=outdir)
    app.build()
    print(app._status.getvalue())
    print(app._warning.getvalue())

    builddir = outdir / '_build' / 'text'
    assert (builddir / 'a.b.c.txt').is_file()
    assert (builddir / 'a.b.e.txt').is_file()
    assert (builddir / 'a.b.x.txt').is_file()

    with open(builddir / 'a.b.c.txt', encoding='utf-8') as f:
        txt = f.read()
    assert 'a.b.c package\n' in txt

    with open(builddir / 'a.b.e.txt', encoding='utf-8') as f:
        txt = f.read()
    assert 'a.b.e.f module\n' in txt

    with open(builddir / 'a.b.x.txt', encoding='utf-8') as f:
        txt = f.read()
    assert 'a.b.x namespace\n' in txt


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-pep420/a',
    options=['--implicit-namespaces', '--separate'],
)
def test_pep_0420_enabled_separate(make_app, apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'a.b.c.rst').is_file()
    assert (outdir / 'a.b.e.rst').is_file()
    assert (outdir / 'a.b.e.f.rst').is_file()
    assert (outdir / 'a.b.x.rst').is_file()
    assert (outdir / 'a.b.x.y.rst').is_file()

    with open(outdir / 'a.b.c.rst', encoding='utf-8') as f:
        rst = f.read()
    assert '.. toctree::\n   :maxdepth: 4\n\n   a.b.c.d\n' in rst

    with open(outdir / 'a.b.e.rst', encoding='utf-8') as f:
        rst = f.read()
    assert '.. toctree::\n   :maxdepth: 4\n\n   a.b.e.f\n' in rst

    with open(outdir / 'a.b.x.rst', encoding='utf-8') as f:
        rst = f.read()
    assert '.. toctree::\n   :maxdepth: 4\n\n   a.b.x.y\n' in rst

    app = make_app('text', srcdir=outdir)
    app.build()
    print(app._status.getvalue())
    print(app._warning.getvalue())

    builddir = outdir / '_build' / 'text'
    assert (builddir / 'a.b.c.txt').is_file()
    assert (builddir / 'a.b.e.txt').is_file()
    assert (builddir / 'a.b.e.f.txt').is_file()
    assert (builddir / 'a.b.x.txt').is_file()
    assert (builddir / 'a.b.x.y.txt').is_file()

    with open(builddir / 'a.b.c.txt', encoding='utf-8') as f:
        txt = f.read()
    assert 'a.b.c package\n' in txt

    with open(builddir / 'a.b.e.f.txt', encoding='utf-8') as f:
        txt = f.read()
    assert 'a.b.e.f module\n' in txt

    with open(builddir / 'a.b.x.txt', encoding='utf-8') as f:
        txt = f.read()
    assert 'a.b.x namespace\n' in txt


@pytest.mark.apidoc(coderoot='test-ext-apidoc-pep420/a')
def test_pep_0420_disabled(make_app, apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert not (outdir / 'a.b.c.rst').exists()
    assert not (outdir / 'a.b.x.rst').exists()

    app = make_app('text', srcdir=outdir)
    app.build()
    print(app._status.getvalue())
    print(app._warning.getvalue())


@pytest.mark.apidoc(coderoot='test-ext-apidoc-pep420/a/b')
def test_pep_0420_disabled_top_level_verify(make_app, apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'c.rst').is_file()
    assert not (outdir / 'x.rst').exists()

    with open(outdir / 'c.rst', encoding='utf-8') as f:
        rst = f.read()
    assert 'c package\n' in rst
    assert 'automodule:: c.d\n' in rst
    assert 'automodule:: c\n' in rst

    app = make_app('text', srcdir=outdir)
    app.build()
    print(app._status.getvalue())
    print(app._warning.getvalue())


@pytest.mark.apidoc(coderoot='test-ext-apidoc-trailing-underscore')
def test_trailing_underscore(make_app, apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'package_.rst').is_file()

    app = make_app('text', srcdir=outdir)
    app.build()
    print(app._status.getvalue())
    print(app._warning.getvalue())

    builddir = outdir / '_build' / 'text'
    with open(builddir / 'package_.txt', encoding='utf-8') as f:
        rst = f.read()
    assert 'package_ package\n' in rst
    assert 'package_.module_ module\n' in rst


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-pep420/a',
    excludes=['b/c/d.py', 'b/e/f.py', 'b/e/__init__.py'],
    options=['--implicit-namespaces', '--separate'],
)
def test_excludes(apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'a.rst').is_file()
    assert (outdir / 'a.b.rst').is_file()
    assert (outdir / 'a.b.c.rst').is_file()  # generated because not empty
    assert not (
        outdir / 'a.b.e.rst'
    ).is_file()  # skipped because of empty after excludes
    assert (outdir / 'a.b.x.rst').is_file()
    assert (outdir / 'a.b.x.y.rst').is_file()


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-pep420/a',
    excludes=['b/e'],
    options=['--implicit-namespaces', '--separate'],
)
def test_excludes_subpackage_should_be_skipped(apidoc):
    """Subpackage exclusion should work."""
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'a.rst').is_file()
    assert (outdir / 'a.b.rst').is_file()
    assert (outdir / 'a.b.c.rst').is_file()  # generated because not empty
    assert not (
        outdir / 'a.b.e.f.rst'
    ).is_file()  # skipped because 'b/e' subpackage is skipped


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-pep420/a',
    excludes=['b/e/f.py'],
    options=['--implicit-namespaces', '--separate'],
)
def test_excludes_module_should_be_skipped(apidoc):
    """Module exclusion should work."""
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'a.rst').is_file()
    assert (outdir / 'a.b.rst').is_file()
    assert (outdir / 'a.b.c.rst').is_file()  # generated because not empty
    assert not (
        outdir / 'a.b.e.f.rst'
    ).is_file()  # skipped because of empty after excludes


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-pep420/a',
    excludes=[],
    options=['--implicit-namespaces', '--separate'],
)
def test_excludes_module_should_not_be_skipped(apidoc):
    """Module should be included if no excludes are used."""
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'a.rst').is_file()
    assert (outdir / 'a.b.rst').is_file()
    assert (outdir / 'a.b.c.rst').is_file()  # generated because not empty
    assert (outdir / 'a.b.e.f.rst').is_file()  # skipped because of empty after excludes


@pytest.mark.apidoc(
    coderoot='test-root',
    options=[
        '--doc-project',
        'プロジェクト名',
        '--doc-author',
        '著者名',
        '--doc-version',
        'バージョン',
        '--doc-release',
        'リリース',
    ],
)
def test_multibyte_parameters(make_app, apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()
    assert (outdir / 'index.rst').is_file()

    conf_py = (outdir / 'conf.py').read_text(encoding='utf8')
    assert "project = 'プロジェクト名'" in conf_py
    assert "author = '著者名'" in conf_py
    assert "version = 'バージョン'" in conf_py
    assert "release = 'リリース'" in conf_py

    app = make_app('text', srcdir=outdir)
    app.build()
    print(app._status.getvalue())
    print(app._warning.getvalue())


@pytest.mark.apidoc(
    coderoot='test-root',
    options=['--ext-mathjax'],
)
def test_extension_parsed(apidoc):
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()

    with open(outdir / 'conf.py', encoding='utf-8') as f:
        rst = f.read()
    assert 'sphinx.ext.mathjax' in rst


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-toc/mypackage',
    options=['--implicit-namespaces'],
)
def test_toc_all_references_should_exist_pep420_enabled(apidoc):
    """All references in toc should exist. This test doesn't say if
    directories with empty __init__.py and and nothing else should be
    skipped, just ensures consistency between what's referenced in the toc
    and what is created. This is the variant with pep420 enabled.
    """
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()

    toc = extract_toc(outdir / 'mypackage.rst')

    refs = [l.strip() for l in toc.splitlines() if l.strip()]
    found_refs = []
    missing_files = []
    for ref in refs:
        if ref and ref[0] in {':', '#'}:
            continue
        found_refs.append(ref)
        filename = f'{ref}.rst'
        if not (outdir / filename).is_file():
            missing_files.append(filename)

    all_missing = ', '.join(missing_files)
    assert len(missing_files) == 0, (
        f'File(s) referenced in TOC not found: {all_missing}\nTOC:\n{toc}'
    )


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-toc/mypackage',
)
def test_toc_all_references_should_exist_pep420_disabled(apidoc):
    """All references in toc should exist. This test doesn't say if
    directories with empty __init__.py and and nothing else should be
    skipped, just ensures consistency between what's referenced in the toc
    and what is created. This is the variant with pep420 disabled.
    """
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()

    toc = extract_toc(outdir / 'mypackage.rst')

    refs = [l.strip() for l in toc.splitlines() if l.strip()]
    found_refs = []
    missing_files = []
    for ref in refs:
        if ref and ref[0] in {':', '#'}:
            continue
        filename = f'{ref}.rst'
        found_refs.append(ref)
        if not (outdir / filename).is_file():
            missing_files.append(filename)

    all_missing = ', '.join(missing_files)
    assert len(missing_files) == 0, (
        f'File(s) referenced in TOC not found: {all_missing}\nTOC:\n{toc}'
    )


def extract_toc(path):
    """Helper: Extract toc section from package rst file"""
    with open(path, encoding='utf-8') as f:
        rst = f.read()

    # Read out the part containing the toctree
    toctree_start = '\n.. toctree::\n'
    toctree_end = '\nSubmodules'

    start_idx = rst.index(toctree_start)
    end_idx = rst.index(toctree_end, start_idx)
    toctree = rst[start_idx + len(toctree_start) : end_idx]

    return toctree


@pytest.mark.apidoc(
    coderoot='test-ext-apidoc-subpackage-in-toc',
    options=['--separate'],
)
def test_subpackage_in_toc(apidoc):
    """Make sure that empty subpackages with non-empty subpackages in them
    are not skipped
    See: https://github.com/sphinx-doc/sphinx/issues/4520
    """
    outdir = apidoc.outdir
    assert (outdir / 'conf.py').is_file()

    assert (outdir / 'parent.rst').is_file()
    with open(outdir / 'parent.rst', encoding='utf-8') as f:
        parent = f.read()
    assert 'parent.child' in parent

    assert (outdir / 'parent.child.rst').is_file()
    with open(outdir / 'parent.child.rst', encoding='utf-8') as f:
        parent_child = f.read()
    assert 'parent.child.foo' in parent_child

    assert (outdir / 'parent.child.foo.rst').is_file()


def test_private(tmp_path):
    (tmp_path / 'hello.py').touch()
    (tmp_path / '_world.py').touch()

    # without --private option
    apidoc_main(['-o', str(tmp_path), str(tmp_path)])
    assert (tmp_path / 'hello.rst').exists()
    assert ':private-members:' not in (tmp_path / 'hello.rst').read_text(
        encoding='utf8'
    )
    assert not (tmp_path / '_world.rst').exists()

    # with --private option
    apidoc_main(['--private', '-f', '-o', str(tmp_path), str(tmp_path)])
    assert (tmp_path / 'hello.rst').exists()
    assert ':private-members:' in (tmp_path / 'hello.rst').read_text(encoding='utf8')
    assert (tmp_path / '_world.rst').exists()


def test_toc_file(tmp_path):
    outdir = tmp_path
    (outdir / 'module').mkdir(parents=True, exist_ok=True)
    (outdir / 'example.py').touch()
    (outdir / 'module' / 'example.py').touch()
    apidoc_main(['-o', str(tmp_path), str(tmp_path)])
    assert (outdir / 'modules.rst').exists()

    content = (outdir / 'modules.rst').read_text(encoding='utf8')
    assert content == (
        'test_toc_file0\n'
        '==============\n'
        '\n'
        '.. toctree::\n'
        '   :maxdepth: 4\n'
        '\n'
        '   example\n'
    )


def test_module_file(tmp_path):
    outdir = tmp_path
    (outdir / 'example.py').touch()
    apidoc_main(['-o', str(tmp_path), str(tmp_path)])
    assert (outdir / 'example.rst').exists()

    content = (outdir / 'example.rst').read_text(encoding='utf8')
    assert content == (
        'example module\n'
        '==============\n'
        '\n'
        '.. automodule:: example\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
    )


def test_module_file_noheadings(tmp_path):
    outdir = tmp_path
    (outdir / 'example.py').touch()
    apidoc_main(['--no-headings', '-o', str(tmp_path), str(tmp_path)])
    assert (outdir / 'example.rst').exists()

    content = (outdir / 'example.rst').read_text(encoding='utf8')
    assert content == (
        '.. automodule:: example\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
    )


def test_package_file(tmp_path):
    outdir = tmp_path
    (outdir / 'testpkg').mkdir(parents=True, exist_ok=True)
    (outdir / 'testpkg' / '__init__.py').touch()
    (outdir / 'testpkg' / 'hello.py').touch()
    (outdir / 'testpkg' / 'world.py').touch()
    (outdir / 'testpkg' / 'subpkg').mkdir(parents=True, exist_ok=True)
    (outdir / 'testpkg' / 'subpkg' / '__init__.py').touch()
    apidoc_main(['-o', str(outdir), str(outdir / 'testpkg')])
    assert (outdir / 'testpkg.rst').exists()
    assert (outdir / 'testpkg.subpkg.rst').exists()

    content = (outdir / 'testpkg.rst').read_text(encoding='utf8')
    assert content == (
        'testpkg package\n'
        '===============\n'
        '\n'
        'Subpackages\n'
        '-----------\n'
        '\n'
        '.. toctree::\n'
        '   :maxdepth: 4\n'
        '\n'
        '   testpkg.subpkg\n'
        '\n'
        'Submodules\n'
        '----------\n'
        '\n'
        'testpkg.hello module\n'
        '--------------------\n'
        '\n'
        '.. automodule:: testpkg.hello\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
        '\n'
        'testpkg.world module\n'
        '--------------------\n'
        '\n'
        '.. automodule:: testpkg.world\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
        '\n'
        'Module contents\n'
        '---------------\n'
        '\n'
        '.. automodule:: testpkg\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
    )

    content = (outdir / 'testpkg.subpkg.rst').read_text(encoding='utf8')
    assert content == (
        'testpkg.subpkg package\n'
        '======================\n'
        '\n'
        'Module contents\n'
        '---------------\n'
        '\n'
        '.. automodule:: testpkg.subpkg\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
    )


def test_package_file_separate(tmp_path):
    outdir = tmp_path
    (outdir / 'testpkg').mkdir(parents=True, exist_ok=True)
    (outdir / 'testpkg' / '__init__.py').touch()
    (outdir / 'testpkg' / 'example.py').touch()
    apidoc_main(['--separate', '-o', str(tmp_path), str(tmp_path / 'testpkg')])
    assert (outdir / 'testpkg.rst').exists()
    assert (outdir / 'testpkg.example.rst').exists()

    content = (outdir / 'testpkg.rst').read_text(encoding='utf8')
    assert content == (
        'testpkg package\n'
        '===============\n'
        '\n'
        'Submodules\n'
        '----------\n'
        '\n'
        '.. toctree::\n'
        '   :maxdepth: 4\n'
        '\n'
        '   testpkg.example\n'
        '\n'
        'Module contents\n'
        '---------------\n'
        '\n'
        '.. automodule:: testpkg\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
    )

    content = (outdir / 'testpkg.example.rst').read_text(encoding='utf8')
    assert content == (
        'testpkg.example module\n'
        '======================\n'
        '\n'
        '.. automodule:: testpkg.example\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
    )


def test_package_file_module_first(tmp_path):
    outdir = tmp_path
    (outdir / 'testpkg').mkdir(parents=True, exist_ok=True)
    (outdir / 'testpkg' / '__init__.py').touch()
    (outdir / 'testpkg' / 'example.py').touch()
    apidoc_main(['--module-first', '-o', str(tmp_path), str(tmp_path)])

    content = (outdir / 'testpkg.rst').read_text(encoding='utf8')
    assert content == (
        'testpkg package\n'
        '===============\n'
        '\n'
        '.. automodule:: testpkg\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
        '\n'
        'Submodules\n'
        '----------\n'
        '\n'
        'testpkg.example module\n'
        '----------------------\n'
        '\n'
        '.. automodule:: testpkg.example\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
    )


def test_package_file_without_submodules(tmp_path):
    outdir = tmp_path
    (outdir / 'testpkg').mkdir(parents=True, exist_ok=True)
    (outdir / 'testpkg' / '__init__.py').touch()
    apidoc_main(['-o', str(tmp_path), str(tmp_path / 'testpkg')])
    assert (outdir / 'testpkg.rst').exists()

    content = (outdir / 'testpkg.rst').read_text(encoding='utf8')
    assert content == (
        'testpkg package\n'
        '===============\n'
        '\n'
        'Module contents\n'
        '---------------\n'
        '\n'
        '.. automodule:: testpkg\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
    )


def test_namespace_package_file(tmp_path):
    outdir = tmp_path
    (outdir / 'testpkg').mkdir(parents=True, exist_ok=True)
    (outdir / 'testpkg' / 'example.py').touch()
    apidoc_main([
        '--implicit-namespace',
        '-o',
        str(tmp_path),
        str(tmp_path / 'testpkg'),
    ])
    assert (outdir / 'testpkg.rst').exists()

    content = (outdir / 'testpkg.rst').read_text(encoding='utf8')
    assert content == (
        'testpkg namespace\n'
        '=================\n'
        '\n'
        '.. py:module:: testpkg\n'
        '\n'
        'Submodules\n'
        '----------\n'
        '\n'
        'testpkg.example module\n'
        '----------------------\n'
        '\n'
        '.. automodule:: testpkg.example\n'
        '   :members:\n'
        '   :show-inheritance:\n'
        '   :undoc-members:\n'
    )


def test_no_duplicates(rootdir, tmp_path):
    """Make sure that a ".pyx" and ".so" don't cause duplicate listings.

    We can't use pytest.mark.apidoc here as we use a different set of arguments
    to apidoc_main
    """
    original_suffixes = sphinx.ext.apidoc._generate.PY_SUFFIXES
    try:
        # Ensure test works on Windows
        sphinx.ext.apidoc._generate.PY_SUFFIXES += ('.so',)

        package = rootdir / 'test-ext-apidoc-duplicates' / 'fish_licence'
        outdir = tmp_path / 'out'
        apidoc_main(['-o', str(outdir), '-T', str(package), '--implicit-namespaces'])

        # Ensure the module has been documented
        assert (outdir / 'fish_licence.rst').is_file()

        # Ensure the submodule only appears once
        text = (outdir / 'fish_licence.rst').read_text(encoding='utf-8')
        count_submodules = text.count(r'fish\_licence.halibut module')
        assert count_submodules == 1

    finally:
        sphinx.ext.apidoc._generate.PY_SUFFIXES = original_suffixes


def test_remove_old_files(tmp_path: Path):
    """Test that old files are removed when using the -r option.

    Also ensure that pre-existing files are not re-written, if unchanged.
    This is required to avoid unnecessary rebuilds.
    """
    module_dir = tmp_path / 'module'
    module_dir.mkdir()
    (module_dir / 'example.py').touch()
    gen_dir = tmp_path / 'gen'
    gen_dir.mkdir()
    (gen_dir / 'other.rst').touch()
    apidoc_main(['-o', str(gen_dir), str(module_dir)])
    assert set(gen_dir.iterdir()) == {
        gen_dir / 'modules.rst',
        gen_dir / 'example.rst',
        gen_dir / 'other.rst',
    }
    example_mtime = (gen_dir / 'example.rst').stat().st_mtime_ns
    apidoc_main(['--remove-old', '-o', str(gen_dir), str(module_dir)])
    assert set(gen_dir.iterdir()) == {gen_dir / 'modules.rst', gen_dir / 'example.rst'}
    assert (gen_dir / 'example.rst').stat().st_mtime_ns == example_mtime


@pytest.mark.sphinx(testroot='ext-apidoc')
def test_sphinx_extension(app: SphinxTestApp) -> None:
    """Test running apidoc as an extension."""
    app.build()
    assert app.warning.getvalue() == ''

    toc_file = app.srcdir / 'generated' / 'modules.rst'
    pkg_file = app.srcdir / 'generated' / 'my_package.rst'
    assert set((app.srcdir / 'generated').iterdir()) == {toc_file, pkg_file}
    modules_content = toc_file.read_text(encoding='utf8')
    assert modules_content == (
        'src\n===\n\n.. toctree::\n   :maxdepth: 3\n\n   my_package\n'
    )
    assert 'show-inheritance' not in pkg_file.read_text(encoding='utf8')
    assert (app.outdir / 'generated' / 'my_package.html').is_file()

    # test a re-build
    app.build()
    assert app.warning.getvalue() == ''

    # TODO check nothing got re-built
    # TODO test that old files are removed



"""Run babel for translations.

Usage:

babel_runner.py extract
    Extract messages from the source code and update the ".pot" template file.

babel_runner.py update
    Update all language catalogues in "sphinx/locale/<language>/LC_MESSAGES"
    with the current messages in the template file.

babel_runner.py compile
    Compile the ".po" catalogue files to ".mo" and ".js" files.
"""

from __future__ import annotations

import json
import logging
import sys
import tempfile
from os import environ
from pathlib import Path

from babel.messages.catalog import Catalog
from babel.messages.extract import (
    DEFAULT_KEYWORDS,
    extract,
    extract_javascript,
    extract_python,
)
from babel.messages.mofile import write_mo
from babel.messages.pofile import read_po, write_po
from babel.util import pathmatch
from jinja2.ext import babel_extract as extract_jinja2

IS_CI = 'CI' in environ
ROOT = Path(__file__).resolve().parent.parent
TEX_DELIMITERS = {
    'variable_start_string': '<%=',
    'variable_end_string': '%>',
    'block_start_string': '<%',
    'block_end_string': '%>',
}
METHOD_MAP = [
    # Extraction from Python source files
    ('**.py', extract_python),
    # Extraction from Jinja2 template files
    ('**/templates/latex/**.tex.jinja', extract_jinja2),
    ('**/templates/latex/**.tex_t', extract_jinja2),
    ('**/templates/latex/**.sty.jinja', extract_jinja2),
    ('**/templates/latex/**.sty_t', extract_jinja2),
    # Extraction from Jinja2 HTML templates
    ('**/themes/**.html', extract_jinja2),
    # Extraction from Jinja2 XML templates
    ('**/themes/**.xml', extract_jinja2),
    # Extraction from JavaScript files
    ('**.js', extract_javascript),
    ('**.js.jinja', extract_javascript),
    ('**.js_t', extract_javascript),
]
OPTIONS_MAP = {
    # Extraction from Python source files
    '**.py': {
        'encoding': 'utf-8',
    },
    # Extraction from Jinja2 template files
    '**/templates/latex/**.tex.jinja': TEX_DELIMITERS.copy(),
    '**/templates/latex/**.tex_t': TEX_DELIMITERS.copy(),
    '**/templates/latex/**.sty.jinja': TEX_DELIMITERS.copy(),
    '**/templates/latex/**.sty_t': TEX_DELIMITERS.copy(),
    # Extraction from Jinja2 HTML templates
    '**/themes/**.html': {
        'encoding': 'utf-8',
        'ignore_tags': 'script,style',
        'include_attrs': 'alt title summary',
    },
}
KEYWORDS = {**DEFAULT_KEYWORDS, '_': None, '__': None}


def run_extract() -> None:
    """Message extraction function."""
    log = _get_logger()

    with open('sphinx/__init__.py', encoding='utf-8') as f:
        lines = f.readlines()
    for line in lines:
        if line.startswith('__version__: Final = '):
            # remove prefix; strip whitespace; remove quotation marks
            sphinx_version = line[21:].strip()[1:-1]
            break

    catalogue = Catalog(project='Sphinx', version=sphinx_version, charset='utf-8')

    base = ROOT / 'sphinx'
    for filename in base.rglob('*'):
        relative_name = filename.relative_to(base)
        for pattern, method in METHOD_MAP:
            if not pathmatch(pattern, str(relative_name)):
                continue

            options = {}
            for opt_pattern, opt_dict in OPTIONS_MAP.items():
                if pathmatch(opt_pattern, str(relative_name)):
                    options = opt_dict
            with open(filename, 'rb') as fileobj:
                for lineno, message, comments, context in extract(
                    method,  # type: ignore[arg-type]
                    fileobj,
                    KEYWORDS,
                    options=options,
                ):
                    catalogue.add(
                        message,
                        None,
                        [(str(relative_name), lineno)],
                        auto_comments=comments,
                        context=context,
                    )
            break

    output_file = ROOT / 'sphinx' / 'locale' / 'sphinx.pot'
    log.info('writing PO template file to %s', output_file.relative_to(ROOT))
    with open(output_file, 'wb') as outfile:
        write_po(outfile, catalogue)


def run_update() -> None:
    """Catalog merging command."""
    log = _get_logger()

    domain = 'sphinx'
    locale_dir = ROOT / 'sphinx' / 'locale'
    template_file = locale_dir / 'sphinx.pot'

    with open(template_file, encoding='utf-8') as infile:
        template = read_po(infile)

    for locale in locale_dir.iterdir():
        filename = locale / 'LC_MESSAGES' / f'{domain}.po'
        if not filename.exists():
            continue

        log.info(
            'updating catalogue %s based on %s',
            filename.relative_to(ROOT),
            template_file.relative_to(ROOT),
        )
        with open(filename, encoding='utf-8') as infile:
            catalogue = read_po(infile, locale=locale.name, domain=domain)

        catalogue.update(template)
        tmp_name = filename.parent / (tempfile.gettempprefix() + filename.name)
        try:
            with open(tmp_name, 'wb') as tmpfile:
                write_po(tmpfile, catalogue)
        except Exception:
            tmp_name.unlink()
            raise

        tmp_name.replace(filename)


def run_compile() -> None:
    """Catalogue compilation command.

    An extended command that writes all message strings that occur in
    JavaScript files to a JavaScript file along with the .mo file.

    Unfortunately, babel's setup command isn't built very extensible, so
    most of the run() code is duplicated here.
    """
    log = _get_logger()

    directory = ROOT / 'sphinx' / 'locale'
    total_errors = {}

    for locale in directory.iterdir():
        po_file = locale / 'LC_MESSAGES' / 'sphinx.po'
        if not po_file.exists():
            continue

        with open(po_file, encoding='utf-8') as infile:
            catalogue = read_po(infile, locale=locale.name)

        if catalogue.fuzzy:
            log.info(
                'catalogue %s is marked as fuzzy, skipping', po_file.relative_to(ROOT)
            )
            continue

        locale_errors = 0
        for message, errors in catalogue.check():
            for error in errors:
                locale_errors += 1
                log.error(
                    'error: %s:%d: %s\nerror:     in message string: %r',
                    po_file.relative_to(ROOT),
                    message.lineno,
                    error,
                    message.string,
                )

        if locale_errors:
            total_errors[locale.name] = locale_errors
            log.info(
                '%d errors encountered in %r locale, skipping',
                locale_errors,
                locale.name,
            )
            continue

        mo_file = locale / 'LC_MESSAGES' / 'sphinx.mo'
        log.info(
            'compiling catalogue %s to %s',
            po_file.relative_to(ROOT),
            mo_file.relative_to(ROOT),
        )
        with open(mo_file, 'wb') as outfile:
            write_mo(outfile, catalogue, use_fuzzy=False)

        js_file = locale / 'LC_MESSAGES' / 'sphinx.js'
        log.info(
            'writing JavaScript strings in catalogue %s to %s',
            po_file.relative_to(ROOT),
            js_file.relative_to(ROOT),
        )
        js_catalogue = {}
        for message in catalogue:
            if any(
                x[0].endswith(('.js', '.js.jinja', '.js_t', '.html'))
                for x in message.locations
            ):
                msgid = message.id
                if isinstance(msgid, list | tuple):
                    msgid = msgid[0]
                js_catalogue[msgid] = message.string

        obj = json.dumps(
            {
                'messages': js_catalogue,
                'plural_expr': catalogue.plural_expr,
                'locale': str(catalogue.locale),
            },
            sort_keys=True,
            indent=4,
        )
        with open(js_file, 'wb') as outfile:
            # to ensure lines end with ``\n`` rather than ``\r\n``:
            outfile.write(f'Documentation.addTranslations({obj});'.encode())

    if total_errors:
        _write_pr_body_line('## Babel catalogue errors')
        _write_pr_body_line('')
        for locale_name, err_count in total_errors.items():
            log.error(
                'error: %d errors encountered in %r locale.', err_count, locale_name
            )
            s = 's' if err_count != 1 else ''
            _write_pr_body_line(f'* {locale_name}: {err_count} error{s}')


def _get_logger() -> logging.Logger:
    log = logging.getLogger('babel')
    log.setLevel(logging.INFO)
    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter('%(message)s'))
    log.addHandler(handler)
    return log


def _write_pr_body_line(message: str) -> None:
    if not IS_CI:
        return
    with open('babel_compile.txt', 'a', encoding='utf-8') as f:
        f.write(f'{message}\n')


if __name__ == '__main__':
    try:
        action = sys.argv[1].lower()
    except IndexError:
        print(__doc__, file=sys.stderr)
        raise SystemExit(2) from None

    if action == 'extract':
        run_extract()
    elif action == 'update':
        run_update()
    elif action == 'compile':
        run_compile()
    elif action == 'all':
        run_extract()
        run_update()
        run_compile()
    else:
        msg = f"invalid action: '{action}'"
        raise ValueError(msg)
    raise SystemExit



"""Test docutils.conf support for several writers."""

from __future__ import annotations

import pytest
from docutils import nodes

from sphinx.testing.util import assert_node
from sphinx.util.docutils import patch_docutils


@pytest.mark.sphinx(
    'dummy',
    testroot='docutilsconf',
    freshenv=True,
)
def test_html_with_default_docutilsconf(app):
    with patch_docutils(app.confdir):
        app.build()

    doctree = app.env.get_doctree('index')
    assert_node(
        doctree[0][1], [nodes.paragraph, ('Sphinx ', [nodes.footnote_reference, '1'])]
    )


@pytest.mark.sphinx(
    'dummy',
    testroot='docutilsconf',
    freshenv=True,
    docutils_conf='[restructuredtext parser]\ntrim_footnote_reference_space: true\n',
    copy_test_root=True,
)
def test_html_with_docutilsconf(app):
    with patch_docutils(app.confdir):
        app.build()

    doctree = app.env.get_doctree('index')
    assert_node(
        doctree[0][1], [nodes.paragraph, ('Sphinx', [nodes.footnote_reference, '1'])]
    )



"""Does foo.svg --> foo.pdf with no change to the file."""

from __future__ import annotations

import shutil
from typing import TYPE_CHECKING

from sphinx.transforms.post_transforms.images import ImageConverter

if TYPE_CHECKING:
    import os

    from sphinx.application import Sphinx
    from sphinx.util.typing import ExtensionMetadata


class MyConverter(ImageConverter):
    conversion_rules = [
        ('image/svg+xml', 'application/pdf'),
    ]

    def is_available(self) -> bool:
        return True

    def convert(
        self, _from: str | os.PathLike[str], _to: str | os.PathLike[str]
    ) -> bool:
        """Mock converts the image from SVG to PDF."""
        shutil.copyfile(_from, _to)
        return True


def setup(app: Sphinx) -> ExtensionMetadata:
    app.add_post_transform(MyConverter)

    return {
        'version': 'builtin',
        'parallel_read_safe': True,
        'parallel_write_safe': True,
    }



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from sphinx.cmd import make_mode
from sphinx.cmd.build import get_parser
from sphinx.cmd.make_mode import run_make_mode

if TYPE_CHECKING:
    from typing import Any

broken_argparse = (
    sys.version_info[:3] <= (3, 12, 6)
    or sys.version_info[:3] == (3, 13, 0)
)  # fmt: skip

DEFAULTS = {
    'filenames': [],
    'jobs': 1,
    'force_all': False,
    'freshenv': False,
    'doctreedir': None,
    'confdir': None,
    'noconfig': False,
    'define': [],
    'htmldefine': [],
    'tags': [],
    'nitpicky': False,
    'verbosity': 0,
    'quiet': False,
    'really_quiet': False,
    'color': 'auto',
    'warnfile': None,
    'warningiserror': False,
    'keep_going': False,
    'traceback': False,
    'pdb': False,
    'exception_on_warning': False,
}

EXPECTED_BUILD_MAIN = {
    'builder': 'html',
    'sourcedir': 'source_dir',
    'outputdir': 'build_dir',
    'filenames': ['filename1', 'filename2'],
    'freshenv': True,
    'noconfig': True,
    'quiet': True,
}

EXPECTED_MAKE_MODE = {
    'builder': 'html',
    'sourcedir': 'source_dir',
    'outputdir': str(Path('build_dir', 'html')),
    'doctreedir': str(Path('build_dir', 'doctrees')),
    'filenames': ['filename1', 'filename2'],
    'freshenv': True,
    'noconfig': True,
    'quiet': True,
}

BUILDER_BUILD_MAIN = [
    '--builder',
    'html',
]
BUILDER_MAKE_MODE = [
    'html',
]
POSITIONAL_DIRS = [
    'source_dir',
    'build_dir',
]
POSITIONAL_FILENAMES = [
    'filename1',
    'filename2',
]
POSITIONAL = POSITIONAL_DIRS + POSITIONAL_FILENAMES
POSITIONAL_MAKE_MODE = BUILDER_MAKE_MODE + POSITIONAL
EARLY_OPTS = [
    '--quiet',
]
LATE_OPTS = [
    '-E',
    '--isolated',
]
OPTS = EARLY_OPTS + LATE_OPTS


def parse_arguments(args: list[str]) -> dict[str, Any]:
    parsed = vars(get_parser().parse_args(args))
    return {k: v for k, v in parsed.items() if k not in DEFAULTS or v != DEFAULTS[k]}


def test_build_main_parse_arguments_pos_first() -> None:
    # <positional...> <opts>
    args = [
        *POSITIONAL,
        *OPTS,
    ]
    assert parse_arguments(args) == EXPECTED_BUILD_MAIN


def test_build_main_parse_arguments_pos_last() -> None:
    # <opts> <positional...>
    args = [
        *OPTS,
        *POSITIONAL,
    ]
    assert parse_arguments(args) == EXPECTED_BUILD_MAIN


def test_build_main_parse_arguments_pos_middle() -> None:
    # <opts> <positional...> <opts>
    args = [
        *EARLY_OPTS,
        *BUILDER_BUILD_MAIN,
        *POSITIONAL,
        *LATE_OPTS,
    ]
    assert parse_arguments(args) == EXPECTED_BUILD_MAIN


@pytest.mark.xfail(
    broken_argparse,
    reason='sphinx-build does not yet support filenames after options',
)
def test_build_main_parse_arguments_filenames_last() -> None:
    args = [
        *POSITIONAL_DIRS,
        *OPTS,
        *POSITIONAL_FILENAMES,
    ]
    assert parse_arguments(args) == EXPECTED_BUILD_MAIN


def test_build_main_parse_arguments_pos_intermixed(
    capsys: pytest.CaptureFixture[str],
) -> None:
    args = [
        *EARLY_OPTS,
        *BUILDER_BUILD_MAIN,
        *POSITIONAL_DIRS,
        *LATE_OPTS,
        *POSITIONAL_FILENAMES,
    ]
    if broken_argparse:
        with pytest.raises(SystemExit):
            parse_arguments(args)
        stderr = capsys.readouterr().err.splitlines()
        assert stderr[-1].endswith('error: unrecognized arguments: filename1 filename2')
    else:
        assert parse_arguments(args) == EXPECTED_BUILD_MAIN


def test_make_mode_parse_arguments_pos_first(monkeypatch: pytest.MonkeyPatch) -> None:
    # -M <positional...> <opts>
    monkeypatch.setattr(make_mode, 'build_main', parse_arguments)
    args = [
        *POSITIONAL_MAKE_MODE,
        *OPTS,
    ]
    assert run_make_mode(args) == EXPECTED_MAKE_MODE


def test_make_mode_parse_arguments_pos_last(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    # -M <opts> <positional...>
    monkeypatch.setattr(make_mode, 'build_main', parse_arguments)
    args = [
        *OPTS,
        *POSITIONAL_MAKE_MODE,
    ]
    with pytest.raises(SystemExit):
        run_make_mode(args)
    stderr = capsys.readouterr().err.splitlines()
    assert stderr[-1].endswith('error: argument --builder/-b: expected one argument')


def test_make_mode_parse_arguments_pos_middle(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    # -M <opts> <positional...> <opts>
    monkeypatch.setattr(make_mode, 'build_main', parse_arguments)
    args = [
        *EARLY_OPTS,
        *POSITIONAL_MAKE_MODE,
        *LATE_OPTS,
    ]
    with pytest.raises(SystemExit):
        run_make_mode(args)
    stderr = capsys.readouterr().err.splitlines()
    assert stderr[-1].endswith('error: argument --builder/-b: expected one argument')


@pytest.mark.xfail(
    broken_argparse,
    reason='sphinx-build does not yet support filenames after options',
)
def test_make_mode_parse_arguments_filenames_last(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    # -M <positional...> <opts> <filenames...>
    monkeypatch.setattr(make_mode, 'build_main', parse_arguments)
    args = [
        *BUILDER_MAKE_MODE,
        *POSITIONAL_DIRS,
        *OPTS,
        *POSITIONAL_FILENAMES,
    ]
    assert run_make_mode(args) == EXPECTED_MAKE_MODE


def test_make_mode_parse_arguments_pos_intermixed(
    monkeypatch: pytest.MonkeyPatch,
    capsys: pytest.CaptureFixture[str],
) -> None:
    # -M <opts> <positional...> <opts> <filenames...>
    monkeypatch.setattr(make_mode, 'build_main', parse_arguments)
    args = [
        *EARLY_OPTS,
        *BUILDER_MAKE_MODE,
        *POSITIONAL_DIRS,
        *LATE_OPTS,
        *POSITIONAL_FILENAMES,
    ]
    with pytest.raises(SystemExit):
        run_make_mode(args)
    stderr = capsys.readouterr().err.splitlines()
    assert stderr[-1].endswith('error: argument --builder/-b: expected one argument')



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from sphinx.testing.util import etree_parse

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from pathlib import Path
    from xml.etree.ElementTree import ElementTree

_etree_cache: dict[Path, ElementTree] = {}


def _parse(path: Path) -> ElementTree:
    if path in _etree_cache:
        return _etree_cache[path]

    _etree_cache[path] = tree = etree_parse(path)
    return tree


@pytest.fixture(scope='package')
def cached_etree_parse() -> Iterator[Callable[[Path], ElementTree]]:
    yield _parse
    _etree_cache.clear()
